import tkinter.ttk as ttk
import warnings
from abc import abstractmethod

import numpy as np
from numpy import ma
from scipy.optimize import minimize

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .style import padding
from .table import ButtonFrame, Table, WrappedLabel

LN_10 = np.log(10)


def stoichiometriesToBoundNames(freeNames, stoichiometries):
    trans = str.maketrans("0123456789", "₀₁₂₃₄₅₆₇₈₉")

    boundNames = []
    for row in stoichiometries:
        boundName = ""
        for freeName, stoichiometry in zip(freeNames, row):
            if stoichiometry == 0:
                continue

            if boundName != "":
                boundName += "·"

            if stoichiometry == 1:
                boundName += freeName
            elif stoichiometry == -1:
                boundName += freeName + "ₙ"
            else:
                boundName += freeName + str(stoichiometry).translate(trans)
        boundNames.append(boundName)
    return np.array(boundNames)


class ComplexSpeciationMixin:
    @property
    def complexIndices(self):
        return ~np.any(self.stoichiometries < 0, 1)

    @property
    def complexCount(self):
        return np.count_nonzero(self.complexIndices)

    @property
    def complexStoichiometries(self):
        return self.stoichiometries[self.complexIndices]

    @property
    def complexBoundNames(self):
        return stoichiometriesToBoundNames(self.freeNames, self.complexStoichiometries)

    complexVariableNames = complexOutputNames = complexBoundNames

    complexOutputStoichiometries = complexStoichiometries

    def complexFreeToBoundConcs(self, freeConcs, complexKs):
        return complexKs * np.prod(freeConcs**self.complexStoichiometries, axis=1)

    def complexObjective(self, free, complexKs, total, M):
        return complexKs @ np.prod(free**M, 1)

    def complexJacobian(self, free, complexKs, total, M):
        return M.T @ (complexKs * np.prod(free**M, 1))

    def complexHessian(self, free, complexKs, total, M):
        return M.T @ (M * np.outer(complexKs * np.prod(free**M, 1), 1 / free))

    def complexGetUpperBounds(self, complexKs, total, M):
        return total


class PolymerSpeciationMixin:
    @property
    def polymerIndices(self):
        return np.any(self.stoichiometries < 0, 1)

    @property
    def polymerCount(self):
        return np.count_nonzero(self.polymerIndices)

    @property
    def polymerStoichiometries(self):
        return self.stoichiometries[self.polymerIndices]

    @property
    def polymerBoundNames(self):
        return stoichiometriesToBoundNames(self.freeNames, self.polymerStoichiometries)

    @property
    def polymerVariableNames(self):
        variableNames = []
        for row, boundName in zip(self.polymerStoichiometries, self.polymerBoundNames):
            if np.count_nonzero(row) == 1:
                variableNames.append(boundName.replace("ₙ", "₂"))
            variableNames.append(boundName)
        return np.array(variableNames)

    @property
    def polymerOutputNames(self):
        return np.ravel(
            [
                [name + " terminal", name + " internal"]
                for name in self.polymerBoundNames
            ]
        )

    @property
    def polymerOutputCount(self):
        return len(self.polymerOutputNames)

    @property
    def polymerOutputStoichiometries(self):
        # stoichiometries in terminal+internal mode: duplicate each polymer
        outputStoichiometries = np.empty([self.polymerOutputCount, self.freeCount])
        M = self.polymerStoichiometries
        outputStoichiometries[::2] = np.where(M > 0, M / 2, abs(M))  # terminal rows
        outputStoichiometries[1::2] = np.where(M > 0, 0, abs(M))  # internal rows
        return outputStoichiometries

    def splitPolymerKs(self, polymerKs):
        k2s = np.zeros(self.freeCount)
        kns = np.zeros(self.freeCount)
        kabs = np.ones(self.polymerCount)
        ks = list(polymerKs)
        for polymerIndex, row in enumerate(self.polymerStoichiometries):
            freeIndex = np.where(row < 0)[0][0]
            if np.count_nonzero(row) == 1:
                k2s[freeIndex], kns[freeIndex] = ks.pop(0), ks.pop(0)
            else:
                kabs[polymerIndex] = ks.pop(0)
        return k2s, kns, kabs

    def getTerminalInternalConcs(self, freeConcs, k2s, kns, kabs):
        terminal = 2 * freeConcs**2 * k2s / (1 - freeConcs * kns)
        internal = freeConcs**3 * k2s * kns / (1 - freeConcs * kns)
        return terminal, internal

    def polymerFreeToBoundConcs(self, freeConcs, k2s, kns, kabs):
        # Treat the terminal and internal units as additional "free" components, so
        # that end-capped polymers can be treated as if they're regular complexes.
        terminal = 2 * freeConcs**2 * k2s / (1 - freeConcs * kns)
        internal = freeConcs**3 * k2s * kns / (1 - freeConcs * kns) ** 2
        componentConcs = np.concatenate([freeConcs, terminal, internal])

        freeCount = self.freeCount
        polymerCount = self.polymerCount

        # split end group (pos) and polymer (neg) stoichiometries
        pos = np.where(self.polymerStoichiometries < 0, 0, self.polymerStoichiometries)
        neg = np.where(
            self.polymerStoichiometries < 0, np.abs(self.polymerStoichiometries), 0
        )

        # each polymer gives two outputs: terminal and internal
        fullStoichiometries = np.zeros([polymerCount * 2, freeCount * 3])

        # terminal and internal states both require end cap concentrations
        fullStoichiometries[::2, :freeCount] = pos
        fullStoichiometries[1::2, :freeCount] = pos
        fullStoichiometries[::2, freeCount : freeCount * 2] = neg
        fullStoichiometries[1::2, freeCount * 2 :] = neg
        return np.repeat(kabs, 2) * np.prod(componentConcs**fullStoichiometries, axis=1)

    def polymerFreeExactSolutionSingle(self, k2, kn, totalSingle):
        roots = np.roots(
            [
                k2 * kn - kn**2,
                totalSingle * kn**2 + 2 * kn - 2 * k2,
                -1 - 2 * totalSingle * kn,
                totalSingle,
            ]
        )
        real_roots = np.real(roots[np.isreal(roots)])
        return np.min(real_roots[real_roots > 0])

    def polymerFreeExactSolution(self, k2s, kns, total):
        return np.array(
            [
                self.polymerFreeExactSolutionSingle(k2, kn, totalSingle)
                for k2, kn, totalSingle in zip(k2s, kns, total)
            ]
        )

    def polymerObjective(self, free, k2s, kns, kabs, total, M):
        if self.polymerCount == 0:
            return 0.0
        pos = np.where(M < 0, 0, M)
        neg = np.where(M < 0, np.abs(M), 0)
        polymerWithoutFactorOfN = free**2 * k2s / (1 - free * kns)

        polymerIntegral = kabs @ np.prod(
            free**pos * polymerWithoutFactorOfN**neg, axis=1
        )
        return np.sum(polymerIntegral)

    def polymerJacobian(self, free, k2s, kns, kabs, total, M):
        if self.polymerCount == 0:
            return np.full(len(free), 0.0)

        pos = np.where(M < 0, 0, M)
        neg = np.where(M < 0, np.abs(M), 0)
        polymerWithFactorOfN = free**2 * k2s * (2 - free * kns) / (1 - free * kns) ** 2
        polymerWithoutFactorOfN = free**2 * k2s / (1 - free * kns)
        polymerConcentration = neg.T @ (
            kabs * np.prod(free**pos * polymerWithFactorOfN**neg, axis=1)
        )
        endCapConcentration = pos.T @ (
            kabs * np.prod(free**pos * polymerWithoutFactorOfN**neg, axis=1)
        )
        return polymerConcentration + endCapConcentration

    def polymerGetUpperBounds(self, k2s, kns, kabs, total, M):
        if self.polymerCount == 0:
            return np.inf
        return np.where(
            self.polymerIndices, self.polymerFreeExactSolution(k2s, kns, total), np.inf
        )


class Speciation(ComplexSpeciationMixin, PolymerSpeciationMixin, moduleFrame.Strategy):
    requiredAttributes = ("stoichiometries",)

    @abstractmethod
    def run(self, variables, totalConcs):
        pass

    # Other modules need to access freeNames and freeCount, but totalConcentrations
    # may not yet be loaded.
    @property
    def freeNames(self):
        try:
            return self.titration.totalConcentrations.freeNames
        except AttributeError:
            return np.array(["Host", "Guest"])

    @property
    def freeCount(self):
        return len(self.freeNames)

    @property
    def boundNames(self):
        return np.append(self.complexBoundNames, self.polymerBoundNames)

    @property
    def boundCount(self):
        return len(self.boundNames)

    @property
    def variableNames(self):
        return np.concatenate([self.complexVariableNames, self.polymerVariableNames])

    @property
    def outputNames(self):
        return np.concatenate(
            [self.freeNames, self.complexOutputNames, self.polymerOutputNames]
        )

    @property
    def outputStoichiometries(self):
        return np.vstack(
            [
                np.eye(self.freeCount),
                self.complexOutputStoichiometries,
                self.polymerOutputStoichiometries,
            ]
        )

    @property
    def formsBinaryComplex(self):
        # polymers include dimers, so treat as a dimer
        M = np.where(self.stoichiometries < 0, 2, self.stoichiometries).tolist()
        result = np.zeros([self.freeCount, self.freeCount], dtype=bool)
        for host in range(self.freeCount):
            for guest in range(self.freeCount):
                target = [0] * self.freeCount
                target[host] += 1
                target[guest] += 1
                result[host, guest] = target in M
        return result

    @property
    def maximumValencyPerGuest(self):
        # polymers can form 2 bonds to themselves, so treat as a trimer
        M = np.where(self.stoichiometries < 0, 3, self.stoichiometries)
        formsBinaryComplex = self.formsBinaryComplex
        result = np.zeros([self.freeCount, self.freeCount], dtype=int)
        for host in range(self.freeCount):
            for guest in range(self.freeCount):
                if not formsBinaryComplex[host, guest]:
                    continue
                elif host == guest:
                    for complex in M:
                        if np.count_nonzero(complex) == 1 and complex[host] != 0:
                            result[host, guest] = max(
                                result[host, guest], complex[host]
                            )
                else:
                    for complex in M:
                        if (
                            np.count_nonzero(complex) == 2
                            and complex[host] == 1
                            and complex[guest] != 0
                        ):
                            result[host, guest] = max(
                                result[host, guest], complex[guest]
                            )
        return result

    def variablesToKs(self, variables):
        complexKs = variables[: self.complexCount]
        polymerKs = variables[self.complexCount :]
        k2s, kns, kabs = self.splitPolymerKs(polymerKs)
        return complexKs, k2s, kns, kabs

    # TODO: rewrite all non-mixin functions to be agnostic to the components of
    # polymerKs, by just working with complexKs and polymerKs, or possibly *polymerKs
    def freeToBoundConcs(self, freeConcs, complexKs, k2s, kns, kabs):
        return np.hstack(
            [
                self.complexFreeToBoundConcs(freeConcs, complexKs),
                self.polymerFreeToBoundConcs(freeConcs, k2s, kns, kabs),
            ]
        )


class SpeciationTable(Table):
    def __init__(self, master, titration):
        sortedTitleLengths = sorted(
            [len(name) for name in titration.totalConcentrations.freeNames]
        )
        self.width = max(6, sortedTitleLengths[-1] + 2)
        self.rowTitleWidth = min(20, sum(sortedTitleLengths[-2:]) + 3 + 2)

        super().__init__(
            master,
            0,
            0,
            titration.totalConcentrations.freeNames,
            rowOptions=("readonlyTitles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
            callback=self.updateTitles,
        )

        if (
            titration.speciation.stoichiometries.shape[1]
            == titration.totalConcentrations.freeCount
        ):
            for boundName, stoichiometry in zip(
                titration.speciation.boundNames, titration.speciation.stoichiometries
            ):
                stoichiometry = stoichiometry.astype(str)
                stoichiometry[stoichiometry == "-1"] = "n"
                self.addRow(boundName, stoichiometry)
        else:
            if titration.totalConcentrations.freeCount == 1:
                self.addRow("", np.array([1]))
            else:
                self.addRow(
                    "",
                    np.concatenate(
                        [
                            [1, 1],
                            [0] * (titration.totalConcentrations.freeCount - 2),
                        ]
                    ).astype(int),
                )
        self.updateTitles()

    def updateTitles(self, *args, **kwargs):
        for cell, title in zip(
            self.cells[2 : self.headerCells :, 1], self.columnTitles
        ):
            cell.set(title)

        self.rowTitles = stoichiometriesToBoundNames(self.columnTitles, self.data)

    def addFreeRow(self, index=-1):
        row = self.cells.shape[0]
        freeRow = np.full(self.cells.shape[1], None)

        freeRow[0] = self.deleteRowButton(row, 0, "Delete Row", state="disabled")
        freeRow[1] = self.readonlyEntry(
            row, 1, self.columnTitles[index], font=self.titleFont
        )

        freeRow[2:] = [
            self.readonlyEntry(row, 2 + column, "0", align="right")
            for column in range(self.cells.shape[1] - 2)
        ]
        freeRow[-1].set("1")

        self.cells = np.insert(self.cells, self.headerCells, freeRow, axis=0)
        self.headerCells += 1

        self.redraw()

    def addColumn(self, firstEntry="", data=None):
        super().addColumn(firstEntry, data)

        for row in range(2, self.headerCells):
            self.cells[row, -1] = self.readonlyEntry(
                row, self.cells.shape[-1], "0", align="right"
            )

        self.addFreeRow()

    def deleteColumn(self, column):
        super().deleteColumn(column)
        self.headerCells -= 1
        self.deleteRow(column)
        self.updateTitles()

    def convertData(self, number):
        if number == "n":
            return -1
        elif number == "":
            return 0
        else:
            return int(number)


class SpeciationPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter speciation table")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        label = WrappedLabel(
            self.frame,
            padding=padding * 2,
            text=(
                "Each column corresponds to a component, and each row to a species.\n"
                "Define each complex by adding a row with the stoichiometry of each"
                " component in the complex. For polymers, use 'n'. Leaving a cell blank"
                " is identical to entering '0'."
            ),
        )
        label.pack(expand=False, fill="both")

        scrolledFrame = ScrolledFrame(
            self.frame, max_width=self.winfo_toplevel().master.winfo_width() - 200
        )
        scrolledFrame.pack(expand=True, fill="both")
        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)

        self.speciationTable = SpeciationTable(self.innerFrame, titration)
        self.speciationTable.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="x", side="bottom")

    def reset(self):
        self.speciationTable.destroy()
        self.speciationTable = SpeciationTable(self.innerFrame, self.titration)
        self.speciationTable.pack(expand=True, fill="both")

    def saveData(self):
        self.stoichiometries = self.speciationTable.data

        self.saved = True
        self.destroy()


class SpeciationDimerisation(Speciation):
    @property
    def stoichiometries(self):
        M = np.array([[2]])
        M.resize([1, self.freeCount])
        return M

    def run(self, variables, totalConcs):
        K = variables[0]
        Htot = totalConcs.T[0]
        H = (-1 + np.sqrt(1 + 8 * Htot * K)) / (4 * K)
        H2 = (1 + 4 * Htot * K - np.sqrt(1 + 8 * Htot * K)) / (8 * K)
        return np.array([H, H2]).T


class SpeciationHG(Speciation):
    @property
    def stoichiometries(self):
        M = np.array([[1, 1]])
        M.resize([1, self.freeCount])
        return M

    def run(self, variables, totalConcs):
        K = variables[0]
        Htot, Gtot = totalConcs.T
        H = (
            np.sqrt(
                Gtot**2 * K**2 - 2 * Gtot * K * (Htot * K - 1) + (Htot * K + 1) ** 2
            )
            - Gtot * K
            + Htot * K
            - 1
        ) / (2 * K)
        G = (
            np.sqrt(
                Htot**2 * K**2 - 2 * Htot * K * (Gtot * K - 1) + (Gtot * K + 1) ** 2
            )
            - Htot * K
            + Gtot * K
            - 1
        ) / (2 * K)
        HG = H * G * K

        return np.array([H, G, HG]).T


# Old, slower speciation algorithm, currently unused. Left in in case results need to
# be compared against it.
class SpeciationCOGS(Speciation):
    def COGS(self, M, y, ks, polymerKs):
        free = y.copy()
        bound = np.empty(len(ks))
        polymers = np.any(M < 0, 1)
        # the index of the species making up each of the polymers
        polymerParents = np.nonzero(M[polymers])[1]
        complexes = ~polymers

        P = np.empty(len(ks))
        P[complexes] = np.sum(M[complexes, :], 1)
        P[polymers] = 2 * ks[polymers] * y[polymerParents]
        P = 1 / max(P)
        tol = 1e-7
        while True:
            bound[complexes] = ks[complexes] * np.prod(free ** M[complexes, :], 1)
            # cap the maximum guess to avoid divergence
            bound[polymers] = np.where(
                free[polymerParents] * polymerKs[polymers] >= 1,
                free[polymerParents] ** 2 * polymerKs[polymers] / P,
                (2 - polymerKs[polymers] * free[polymerParents])
                * (ks[polymers] * free[polymerParents] ** 2)
                / ((1 - polymerKs[polymers] * free[polymerParents]) ** 2),
            )
            total = free + abs(M.T) @ bound  # total concentrations of species
            if all((total - y) <= tol * y):
                break
            # to handle 0 total concentration
            invTotal = np.where(total == 0, 1, 1 / total)
            free *= (y * invTotal) ** P

        # For polymers, return separate entries for terminal and internal groups
        bound = []
        paddedPolymerParents = ma.array(np.empty(len(ks)), mask=True)
        paddedPolymerParents[polymers] = polymerParents
        for k, polymerK, stoichiometries, isPolymer, parent in zip(
            ks, polymerKs, M, self.polymerIndices, paddedPolymerParents
        ):
            if not isPolymer:
                bound.append(k * np.prod(free**stoichiometries))
            else:
                # terminal
                bound.append(2 * free[parent] ** 2 * k / (1 - free[parent] * polymerK))
                # internal
                bound.append(
                    (free[parent] ** 3 * k * polymerK)
                    / (1 - free[parent] * polymerK) ** 2
                )

        return free, np.array(bound)

    def run(self, variables, totalConcs):
        ks, polymerKs = self.variablesToKs(variables)
        numPoints = totalConcs.shape[0]
        free = np.zeros((numPoints, self.titration.totalConcentrations.freeCount))
        bound = np.zeros((numPoints, self.complexCount + 2 * self.polymerCount))
        for i in range(numPoints):
            free[i], bound[i] = self.COGS(
                self.stoichiometries, totalConcs[i], ks, polymerKs
            )
        return free, bound


class SpeciationSolver(Speciation):
    # Optimise f(total * log10(free))
    # This makes the derivative equal to ln(10) * (free + bound - total) / total
    #
    # So gtol in the minimisation algorithm can be set to the desired precision in
    # the total concentrations.
    def objective(
        self, logFreeTimesTotal, complexKs, k2s, kns, kabs, total, complexM, polymerM
    ):
        free = 10 ** (logFreeTimesTotal / total)
        return (
            np.sum(free)
            - np.sum(logFreeTimesTotal) * LN_10
            + self.complexObjective(free, complexKs, total, complexM)
            + self.polymerObjective(free, k2s, kns, kabs, total, polymerM)
        )

    def jacobian(
        self, logFreeTimesTotal, complexKs, k2s, kns, kabs, total, complexM, polymerM
    ):
        free = 10 ** (logFreeTimesTotal / total)
        return (
            (
                free
                + self.complexJacobian(free, complexKs, total, complexM)
                + self.polymerJacobian(free, k2s, kns, kabs, total, polymerM)
                - total
            )
            * LN_10
            / total
        )

    def objectiveScaled(
        self, logFreeTimesTotal, complexKs, k2s, kns, kabs, total, complexM, polymerM
    ):
        free = 10 ** (logFreeTimesTotal / self.scaling_factor / total)
        return (
            np.sum(free)
            - np.sum(logFreeTimesTotal / self.scaling_factor) * LN_10
            + self.complexObjective(free, complexKs, total, complexM)
            + self.polymerObjective(free, k2s, kns, kabs, total, polymerM)
        ) * self.scaling_factor

    def jacobianScaled(
        self, logFreeTimesTotal, complexKs, k2s, kns, kabs, total, complexM, polymerM
    ):
        free = 10 ** (logFreeTimesTotal / self.scaling_factor / total)
        return (
            (
                free
                + self.complexJacobian(free, complexKs, total, complexM)
                + self.polymerJacobian(free, k2s, kns, kabs, total, polymerM)
                - total
            )
            * LN_10
            / total
        )

    def smoothObjective(self, logFreeTimesTotal, *args, **kwargs):
        upperBounds = self.getDomainUpperBounds(*args, **kwargs)
        if np.all(logFreeTimesTotal <= upperBounds):
            return self.objective(logFreeTimesTotal, *args, **kwargs)
        print("used truncation in objective")

        truncatedX = np.clip(logFreeTimesTotal, None, upperBounds)
        return self.objective(truncatedX, *args, **kwargs) + np.sum(
            (logFreeTimesTotal - truncatedX)
            * self.jacobian(truncatedX, *args, **kwargs)
        )

    def smoothJacobian(self, logFreeTimesTotal, *args, **kwargs):
        upperBounds = self.getDomainUpperBounds(*args, **kwargs)
        if np.all(logFreeTimesTotal <= upperBounds):
            return self.jacobian(logFreeTimesTotal, *args, **kwargs)
        print("used truncation in jacobian")

        truncatedX = np.clip(logFreeTimesTotal, None, upperBounds)
        return self.jacobian(truncatedX, *args, **kwargs)

    def getDomainUpperBounds(
        self, complexKs, k2s, kns, kabs, total, complexM, polymerM
    ):
        return total * np.log10(
            self.polymerGetUpperBounds(k2s, kns, kabs, total, polymerM),
        )

    # Could be refined iteratively, by computing the LB using this method, then taking
    # UB = self.freeToBoundConcs(free=LB), calculating a new LB using that UB, etc.
    def getUpperBounds(self, complexKs, k2s, kns, kabs, total, complexM, polymerM):
        return total * np.log10(
            np.minimum(
                self.complexGetUpperBounds(complexKs, total, complexM),
                self.polymerGetUpperBounds(k2s, kns, kabs, total, polymerM),
            )
        )

    def getLowerBounds(self, complexKs, k2s, kns, kabs, total, complexM, polymerM):
        maxFree = np.minimum(
            self.complexGetUpperBounds(complexKs, total, complexM),
            self.polymerGetUpperBounds(k2s, kns, kabs, total, polymerM),
        )
        return total * np.log10(
            (total * maxFree)
            / (
                maxFree
                + self.complexJacobian(maxFree, complexKs, total, complexM)
                + self.polymerJacobian(maxFree, k2s, kns, kabs, total, polymerM)
            )
        )

    def run(self, variables, totalConcs):
        complexKs, k2s, kns, kabs = self.variablesToKs(variables)
        numPoints = totalConcs.shape[0]

        free = np.empty((numPoints, self.freeCount))
        bound = np.empty((numPoints, self.outputCount - self.freeCount))

        for i in range(numPoints):
            additionTotalConcs = totalConcs[i]

            # Filter the to exclude species and complexes that will have a concentration
            # of 0
            zeroFree = additionTotalConcs == 0
            zeroBound = np.any(self.stoichiometries[:, zeroFree], axis=1)
            if all(zeroBound):
                free[i] = additionTotalConcs
                bound[i] = 0
                continue
            zeroComplexes = zeroBound[~self.polymerIndices]
            zeroPolymers = zeroBound[self.polymerIndices]

            filteredKs = complexKs[~zeroComplexes]
            filteredK2s = k2s[~zeroFree]
            filteredKns = kns[~zeroFree]
            filteredKabs = kabs[~zeroPolymers]

            filteredTotal = additionTotalConcs[~zeroFree]
            filteredComplexM = self.complexStoichiometries[~zeroComplexes, :][
                :, ~zeroFree
            ]
            filteredPolymerM = self.polymerStoichiometries[~zeroPolymers, :][
                :, ~zeroFree
            ]

            args = (
                filteredKs,
                filteredK2s,
                filteredKns,
                filteredKabs,
                filteredTotal,
                filteredComplexM,
                filteredPolymerM,
            )

            lb = self.getLowerBounds(*args)
            ub = self.getUpperBounds(*args)
            if any(lb > ub):
                # Correct for rounding errors. Unsure if this is necessary, as the only
                # obvious case when it should happen is if no complexes are formed,
                # which should be caught above.
                mask = np.logical_and(lb > ub, np.isclose(lb, ub))
                lb[mask], ub[mask] = ub[mask], lb[mask]

            # TODO: deal with cases where lb and ub are very close together!
            # self.scaling_factor = 1000 / min(ub - lb)
            self.scaling_factor = 1000 / np.min(
                np.abs(filteredTotal * np.log10(filteredTotal))
            )
            self.scaling_factor = 1

            if i == 0:
                # Initial guess: all species 100% free, only polymers are formed
                x0 = ub
            else:
                # If the total concentration increased, the initial guess is that all
                # the added molecules are free.
                # If the total concentration decreased, the initial guess is that the
                # free concentration decreases by the same fraction.
                initialGuess = free[i - 1].copy()
                difference = additionTotalConcs - totalConcs[i - 1]
                concsIncreased = difference >= 0

                initialGuess[concsIncreased] += difference[concsIncreased]
                initialGuess[~concsIncreased] *= (
                    totalConcs[i][~concsIncreased] / totalConcs[i - 1][~concsIncreased]
                )
                x0 = filteredTotal * np.log10(initialGuess[~zeroFree])
                x0 = np.clip(x0, lb, ub)

            result = minimize(
                self.objectiveScaled,
                jac=self.jacobianScaled,
                args=args,
                x0=x0 * self.scaling_factor,
                bounds=np.vstack([lb, ub]).T * self.scaling_factor,
                method="L-BFGS-B",
                options={
                    "ftol": 0.0,
                    "gtol": 1e-6 * LN_10,
                },
            )
            if max(abs(result.jac)) > 1e-6 * LN_10:
                self.scaling_factor *= 10_000
                improvedResult = minimize(
                    self.objectiveScaled,
                    jac=self.jacobianScaled,
                    args=args,
                    x0=result.x,
                    bounds=np.vstack([lb, ub]).T * self.scaling_factor,
                    method="L-BFGS-B",
                    options={
                        "ftol": 0.0,
                        "gtol": 1e-6 * LN_10,
                    },
                )
                if max(abs(improvedResult.jac)) < max(abs(result.jac)):
                    result = improvedResult
                else:
                    warnings.warn(
                        "Desired accuracy not achieved in speciation",
                        RuntimeWarning,
                    )

            logFree = result.x / self.scaling_factor / filteredTotal

            free[i, ~zeroFree] = 10**logFree
            free[i, zeroFree] = 0
            # get the concentrations of the bound species from those of the free
            bound[i] = self.freeToBoundConcs(free[i], complexKs, k2s, kns, kabs)

        return np.hstack([free, bound])


class SpeciationHG2(SpeciationSolver):
    @property
    def stoichiometries(self):
        if self.freeCount < 2:
            return np.array([[1] * self.freeCount])
        else:
            return np.pad(
                array=[[1, 1], [1, 2]],
                pad_width=[[0, 0], [0, self.freeCount - 2]],
                mode="constant",
                constant_values=0,
            )


class SpeciationCustom(SpeciationSolver):
    Popup = SpeciationPopup
    popupAttributes = ("stoichiometries",)


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Equilibria"
    dropdownLabelText = "Select a binding isotherm:"
    dropdownOptions = {
        "1:1 binding": SpeciationHG,
        "1:2 binding": SpeciationHG2,
        "Dimerisation": SpeciationDimerisation,
        "Custom": SpeciationCustom,
        # "Custom Grad": SpeciationCustomGrad,
    }
    attributeName = "speciation"



import functools
import tkinter as tk
import tkinter.ttk as ttk

import matplotlib
from matplotlib import cbook, offsetbox
from matplotlib.axes._base import _AxesBase
from matplotlib.backend_bases import NavigationToolbar2, _Mode, cursors
from matplotlib.backends._backend_tk import (
    FigureCanvasTk,
    NavigationToolbar2Tk,
    ToolTip,
    cursord,
)

matplotlib.use("TkAgg")


class NavigationToolbarVertical(NavigationToolbar2Tk):
    def __init__(self, canvas, window=None, *, pack_toolbar=True):
        """
        Parameters
        ----------
        canvas : `FigureCanvas`
            The figure canvas on which to operate.
        window : tk.Window
            The tk.Window which owns this toolbar.
        pack_toolbar : bool, default: True
            If True, add the toolbar to the parent's pack manager's packing
            list during initialization with ``side="bottom"`` and ``fill="x"``.
            If you want to use the toolbar with a different layout manager, use
            ``pack_toolbar=False``.
        """

        if window is None:
            window = canvas.get_tk_widget().master
        tk.Frame.__init__(
            self,
            master=window,
            borderwidth=2,
            width=int(canvas.figure.bbox.width),
            height=50,
        )

        self._buttons = {}
        for text, tooltip_text, image_file, callback in self.toolitems:
            if callback == "configure_subplots":
                # configure subplots doesn't work with constrained layout
                continue
            if text is None:
                # Add a spacer; return value is unused.
                self._Spacer()
            else:
                self._buttons[text] = button = self._Button(
                    text,
                    str(cbook._get_data_path(f"images/{image_file}.png")),
                    toggle=callback in ["zoom", "pan"],
                    command=getattr(self, callback),
                )
                if tooltip_text is not None:
                    ToolTip.createToolTip(button, tooltip_text)

        self.message = tk.StringVar(master=self)
        self.message.set("\N{NO-BREAK SPACE}\n\N{NO-BREAK SPACE}")
        self._message_label = ttk.Label(
            master=self,
            textvariable=self.message,
            width=len("y=-1234.5"),
        )
        self._message_label.bind(
            "<Configure>",
            lambda *args: self._message_label.configure(width=len("y=-1234.5")),
        )
        self._message_label.pack(side=tk.TOP)

        NavigationToolbar2.__init__(self, canvas)

    def set_message(self, s):
        self.message.set(s.replace(" ", "\n"))


# updates the navigation toolbar to use ttk, and the correct ttkbootstrap style
def _Button(self, text, image_file, toggle, command):
    if tk.TkVersion >= 8.6:
        PhotoImage = tk.PhotoImage
    else:
        from PIL.ImageTk import PhotoImage
    image = PhotoImage(master=self, file=image_file) if image_file is not None else None
    if not toggle:
        b = ttk.Button(
            master=self,
            text=text,
            image=image,
            command=command,
            style="Outline.TButton",
        )
    else:
        var = tk.IntVar(master=self)
        b = ttk.Checkbutton(
            master=self,
            text=text,
            image=image,
            command=command,
            variable=var,
            style="Outline.Toolbutton",
        )
        b.var = var
    b._ntimage = image
    if isinstance(self, NavigationToolbarVertical):
        b.pack(side=tk.TOP, pady=2)
    else:
        b.pack(side=tk.LEFT)
    return b


def _Spacer(self):
    if isinstance(self, NavigationToolbarVertical):
        s = ttk.Separator(master=self, orient=tk.HORIZONTAL)
        s.pack(side=tk.TOP, pady="2")
    else:
        s = ttk.Separator(master=self, orient=tk.VERTICAL)
        s.pack(side=tk.LEFT, padx="3p")
    return s


def _update_buttons_checked(self):
    # sync button checkstates to match active mode
    for text, mode in [("Zoom", _Mode.ZOOM), ("Pan", _Mode.PAN)]:
        if text in self._buttons:
            if self.mode == mode:
                self._buttons[text].var.set(1)
            else:
                self._buttons[text].var.set(0)


def nop(self):
    pass


def __init__(self, ref_artist, use_blit=False):
    self.ref_artist = ref_artist
    if not ref_artist.pickable():
        ref_artist.set_picker(True)
    self.got_artist = False
    self._hover = False
    self._use_blit = use_blit and self.canvas.supports_blit
    callbacks = ref_artist.figure._canvas_callbacks
    self._disconnectors = [
        functools.partial(
            callbacks.disconnect, callbacks._connect_picklable(name, func)
        )
        for name, func in [
            ("pick_event", self.on_pick),
            ("button_release_event", self.on_release),
            ("motion_notify_event", self.on_motion),
        ]
    ]


def on_motion(self, evt):
    # Only check if the widget lock is available, setting it would prevent
    # picking.
    if not (self._check_still_parented() and self.canvas.widgetlock.available(self)):
        return

    picker = self.ref_artist.get_picker()
    if callable(picker):
        inside, _ = picker(self, evt)
    else:
        inside, _ = self.ref_artist.contains(evt)

    # If the mouse is moving quickly while dragging, it may leave the artist,
    # but should still use the move cursor.
    if inside or self.got_artist:
        # Ideally, this should use an open hand cursor on hover, and a closed
        # hand when dragging, but those cursors are not natively provided by
        # all platforms.
        self._hover = True
        self.canvas.set_cursor(cursors.MOVE)
    elif self._hover:
        # Only change the cursor back if this is the widget that set it, to
        # avoid multiple draggable widgets fighting over the cursor.
        self._hover = False
        self.canvas.set_cursor(cursors.POINTER)

    if self._check_still_parented() and self.got_artist:
        dx = evt.x - self.mouse_x
        dy = evt.y - self.mouse_y
        self.update_offset(dx, dy)
        if self._use_blit:
            self.canvas.restore_region(self.background)
            self.ref_artist.draw(self.ref_artist.figure._get_renderer())
            self.canvas.blit()
        else:
            self.canvas.draw()


def on_pick(self, evt):
    if self._check_still_parented() and evt.artist == self.ref_artist:
        self.mouse_x = evt.mouseevent.x
        self.mouse_y = evt.mouseevent.y
        self.got_artist = True
        if self._use_blit:
            self.ref_artist.set_animated(True)
            self.canvas.draw()
            self.background = self.canvas.copy_from_bbox(self.ref_artist.figure.bbox)
            self.ref_artist.draw(self.ref_artist.figure._get_renderer())
            self.canvas.blit()
        self.save_offset()


def on_release(self, event):
    if self._check_still_parented() and self.got_artist:
        self.finalize_offset()
        self.got_artist = False

        if self._use_blit:
            self.ref_artist.set_animated(False)


def disconnect(self):
    """Disconnect the callbacks."""
    for disconnector in self._disconnectors:
        disconnector()
    if self._hover:
        self.canvas.set_cursor(cursors.POINTER)


original_scroll_event_windows = FigureCanvasTk.scroll_event_windows


def scroll_event_windows(self, event):
    # gives an error when scrolling over "tk busy" otherwise
    if type(event.widget) is not str:
        original_scroll_event_windows(self, event)


original_clear = _AxesBase.clear


def clear(self, *args, **kwargs):
    if hasattr(self, "legend_") and self.legend_ is not None:
        self.legend_.axes = self.legend_.figure = None
    original_clear(self, *args, **kwargs)


def applyPatch():
    # makes buttons use ttk widgets
    NavigationToolbar2Tk._Button = _Button
    NavigationToolbar2Tk._Spacer = _Spacer
    NavigationToolbar2Tk._update_buttons_checked = _update_buttons_checked

    # implements PR #25412, and #25442 below mpl version 3.7.2
    offsetbox.DraggableBase.__init__ = __init__
    offsetbox.DraggableBase.cids = property(
        lambda self: [disconnect.args[0] for disconnect in self._disconnectors[:2]]
    )
    offsetbox.DraggableBase.on_motion = on_motion
    offsetbox.DraggableBase.on_pick = on_pick
    offsetbox.DraggableBase.on_release = on_release
    offsetbox.DraggableBase.disconnect = disconnect
    _AxesBase.clear = clear

    # implements PR #25413
    cursord[cursors.SELECT_REGION] = "crosshair"

    FigureCanvasTk.scroll_event_windows = scroll_event_windows



import tkinter as tk
import tkinter.messagebox as mb
import tkinter.ttk as ttk

import numpy as np
from numpy import ma

from . import style
from .style import cellWidth, padding


class Table(ttk.Frame):
    width = cellWidth
    rowTitleWidth = None

    titleFont = None
    italicFont = None
    fitToText = "New column"

    def __init__(
        self,
        master,
        headerGridRows,
        headerCells,
        columnTitles=[],
        *,
        maskBlanks=False,
        blankValue="?",
        rowOptions=[],
        columnOptions=[],
        boldTitles=False,
        callback=None,
        allowGuesses=False,
        **kwargs,
    ):
        self.rowOptions = rowOptions
        self.columnOptions = columnOptions
        self.maskBlanks = maskBlanks
        self.blankValue = blankValue
        self.allowGuesses = allowGuesses
        super().__init__(master, padding=padding, **kwargs)
        if callback is not None:
            callback = self.register(callback)
        self.callback = callback

        if Table.titleFont is None:
            Table.titleFont = style.boldFont
        if not boldTitles:
            self.titleFont = "TkTextFont"
        if Table.italicFont is None:
            Table.italicFont = style.italicFont

        # Grid rows at the top that will be left empty for a subclass to fill
        # in with other widgets.
        self.headerGridRows = headerGridRows
        # Rows at the top of the table that won't be used for data, for the
        # subclass to fill in with widgets that should be treated as cells.
        # Two rows are reserved for column titles and delete buttons.
        self.headerCells = headerCells + 2

        if "new" in rowOptions:
            self.newRowButton = self.button(
                1,
                0,
                "New row",
                self.newRow,
                style="success.Outline.TButton",
            )
        if "new" in columnOptions:
            self.newColumnButton = self.button(
                0,
                1,
                "New column",
                self.newColumn,
                style="success.Outline.TButton",
            )

        self.initEmptyCells()
        self.columnTitles = columnTitles

    def initEmptyCells(self):
        self.cells = np.full((self.headerCells, 2), None)

    # all widget creation methods use cell rows, not grid rows, as arguments

    def entry(
        self,
        row,
        column,
        text="",
        align="left",
        columnspan=1,
        *,
        addCallback=True,
        width=None,
        **kwargs,
    ):
        entry = ttk.Entry(
            self,
            width=width or self.width * columnspan,
            justify=align,
            **kwargs,
        )

        entry.stringVar = tk.StringVar(entry)
        entry.configure(textvariable=entry.stringVar)
        if self.callback is not None and addCallback:
            # Callback is registered once for the entire table, but StringVar.trace_add
            # only supports python functions.
            self.tk.eval(
                f"trace add variable {entry.stringVar._name} write {self.callback}"
            )
        if self.allowGuesses:
            entry.stringVar.trace_add("write", lambda *args: self.checkIfGuess(entry))

        entry.get = entry.stringVar.get
        entry.set = entry.stringVar.set

        entry.set(text)
        entry.grid(
            row=self.headerGridRows + row,
            column=column,
            sticky="nesw",
            columnspan=columnspan,
        )

        entry.bind("<<Paste>>", lambda *args: self.paste(entry))
        entry.bind("<Tab>", lambda *args: self.tab(entry))
        entry.bind("<Shift-Tab>", lambda *args: self.shiftTab(entry))

        # Left and Right move the cursor inside the entry
        directions = {"Up": (-1, 0), "Down": (1, 0)}
        for key, value in directions.items():
            entry.bind(
                f"<{key}>",
                (lambda direction: lambda *args: self.arrowKey(entry, direction))(
                    value
                ),
            )

        return entry

    def checkIfGuess(self, entry):
        if entry.get().startswith("~"):
            entry.configure(style="info.TEntry")
            entry.configure(font=self.italicFont)
        elif entry.configure("style")[-1] == "info.TEntry":
            # Only reset the style if it was set to info.TEntry
            entry.configure(style="primary.TEntry")
            entry.configure(font="TkTextFont")

    def readonlyEntry(self, *args, **kwargs):
        entry = self.entry(*args, style="TLabel", addCallback=False, **kwargs)
        entry.state(["readonly"])
        entry.grid(padx=5, pady=5)
        entry.configure(takefocus=False)
        return entry

    def label(self, row, column, text="", columnspan=1, **kwargs):
        frame = ttk.Frame(self, width=self.width * columnspan, borderwidth=5)
        frame.grid(
            row=self.headerGridRows + row,
            column=column,
            sticky="nesw",
            columnspan=columnspan,
        )
        label = ttk.Label(frame, text=text, **kwargs)
        label.pack()
        return frame

    def button(
        self,
        row,
        column,
        text="",
        command=None,
        columnspan=1,
        style="Outline.TButton",
        **kwargs,
    ):
        button = ttk.Button(
            self, text=text, command=command, style=style, takefocus=False, **kwargs
        )
        button.grid(
            row=self.headerGridRows + row,
            column=column,
            sticky="nesw",
            columnspan=columnspan,
            padx=1,
            pady=1,
        )
        return button

    def deleteRowButton(self, *args, **kwargs):
        button = self.button(*args, style="danger.Outline.TButton", **kwargs)
        button.configure(
            command=lambda: self.deleteRow(
                button.grid_info()["row"] - self.headerGridRows
            )
        )
        return button

    def deleteColumnButton(self, *args, **kwargs):
        button = self.button(*args, style="danger.Outline.TButton", **kwargs)
        button.configure(
            command=lambda button=button: self.deleteColumn(
                button.grid_info()["column"]
            )
        )
        return button

    def dropdown(self, row, column, choices=[], default=None, columnspan=1):
        stringVar = tk.StringVar()
        if default is None:
            default = choices[0]
        optionMenu = ttk.OptionMenu(
            self, stringVar, default, *choices, style="Outline.TMenubutton"
        )
        optionMenu.configure(takefocus=0)
        optionMenu.grid(
            row=self.headerGridRows + row,
            column=column,
            sticky="nesw",
            columnspan=columnspan,
            padx=1,
            pady=1,
        )
        return optionMenu, stringVar

    def redraw(self):
        for (row, column), cell in np.ndenumerate(self.cells):
            if cell is not None:
                cell.grid(row=self.headerGridRows + row, column=column)

    def addRow(self, firstEntry="", data=None):
        row = self.cells.shape[0]
        newRow = np.full(self.cells.shape[1], None)
        rowTitleWidth = self.rowTitleWidth or self.width
        if "delete" in self.rowOptions:
            newRow[0] = self.deleteRowButton(row, 0, "Delete Row")
        if "readonlyTitles" in self.rowOptions:
            newRow[1] = self.readonlyEntry(
                row, 1, firstEntry, font=self.titleFont, width=rowTitleWidth
            )
        elif "titles" in self.rowOptions:
            newRow[1] = self.entry(
                row, 1, firstEntry, font=self.titleFont, width=rowTitleWidth
            )
        for column in range(self.cells.shape[1] - 2):
            entry = self.entry(row, 2 + column, align="right")
            if data is not None:
                entry.set(data[column])
            newRow[2 + column] = entry

        self.cells = np.vstack((self.cells, newRow))

    def addColumn(self, firstEntry="", data=None):
        column = self.cells.shape[1]
        newColumn = np.full(self.cells.shape[0], None)
        if "delete" in self.columnOptions:
            newColumn[0] = self.deleteColumnButton(0, column, "Delete Column")
        if "readonlyTitles" in self.columnOptions:
            newColumn[1] = self.readonlyEntry(
                1,
                column,
                firstEntry,
                align="center",
                font=self.titleFont,
            )
        elif "titles" in self.columnOptions:
            newColumn[1] = self.entry(1, column, firstEntry, font=self.titleFont)
        for row in range(self.headerCells, self.cells.shape[0]):
            entry = self.entry(row, column, align="right")
            if data is not None:
                entry.set(data[row - self.headerCells])
            newColumn[row] = entry

        self.cells = np.hstack((self.cells, newColumn[:, None]))

    # So that subclasses can specify default values when new rows/columns are added
    # by a button press.
    def newColumn(self):
        self.addColumn()

    def newRow(self):
        self.addRow()

    def deleteRow(self, row):
        for element in self.cells[row]:
            if element is not None:
                element.destroy()
        self.cells = np.delete(self.cells, row, 0)
        self.redraw()

    def deleteColumn(self, column):
        for element in self.cells[:, column]:
            if element is not None:
                element.destroy()
        self.cells = np.delete(self.cells, column, 1)
        self.redraw()

    def resetData(self):
        for row in self.cells:
            for element in row:
                if element is not None:
                    element.destroy()
        self.initEmptyCells()

    def convertData(self, number):
        if self.maskBlanks and (
            number == self.blankValue or (self.allowGuesses and number.startswith("~"))
        ):
            return np.nan
        elif number == "":
            if self.maskBlanks:
                raise ValueError(
                    "Please enter a value in each cell. To optimise a variable, enter"
                    ' "?".'
                )
            raise ValueError("Please enter a value in each cell.")
        return float(number)

    def setFocus(self, widget):
        widget.tk.eval(f"tk::TabToWindow {widget._w}")

    def tab(self, entry):
        info = entry.grid_info()
        row, column = info["row"] - self.headerGridRows, info["column"]
        rows, columns = self.cells.shape
        while not (row == rows - 1 and column == columns - 1):
            column += 1
            if column >= columns:
                column = 0
                row += 1
            nextCell = self.cells[row, column]
            if isinstance(nextCell, tk.Widget) and nextCell.tk.call(
                "::tk::FocusOK", nextCell
            ):
                self.setFocus(nextCell)
                return "break"

    def shiftTab(self, entry):
        info = entry.grid_info()
        row, column = info["row"] - self.headerGridRows, info["column"]
        rows, columns = self.cells.shape
        while not (row == 0 and column == 0):
            column -= 1
            if column < 0:
                column = columns - 1
                row -= 1
            prevCell = self.cells[row, column]
            if isinstance(prevCell, tk.Widget) and prevCell.tk.call(
                "::tk::FocusOK", prevCell
            ):
                self.setFocus(prevCell)
                return "break"

    def arrowKey(self, entry, direction):
        info = entry.grid_info()
        row, column = info["row"] - self.headerGridRows, info["column"]
        rows, columns = self.cells.shape
        targetRow, targetColumn = row + direction[0], column + direction[1]
        if (
            targetRow < 0
            or targetColumn < 0
            or targetRow >= rows
            or targetColumn >= columns
        ):
            return

        targetCell = self.cells[targetRow, targetColumn]
        if isinstance(targetCell, tk.Widget) and targetCell.tk.call(
            "::tk::FocusOK", targetCell
        ):
            self.setFocus(targetCell)
            return "break"

    def paste(self, entry):
        try:
            pasteContent = self.clipboard_get()
        except tk.TclError:
            # No text in clipboard
            return
        if "\n" not in pasteContent and "\r" not in pasteContent:
            # Handle as a regular paste. Copying a single cell or row from Excel should
            # still end the string with a newline.
            return

        info = entry.grid_info()
        row, column = info["row"] - self.headerGridRows, info["column"]
        rows, columns = self.cells[row:, column:].shape
        lines = pasteContent.splitlines()
        if len(lines) > rows:
            mb.showerror(
                title="Paste error",
                message=f"Cannot paste {len(lines)} lines into {rows} rows.",
                parent=self,
            )
            return "break"
        for rowOffset, line in enumerate(lines):
            for columnOffset, value in enumerate(line.split("\t", columns - 1)):
                self.cells[row + rowOffset, column + columnOffset].set(value)
        self.setFocus(self.cells[row + rowOffset, column + columnOffset])
        return "break"

    @property
    def dataCells(self):
        return self.cells[self.headerCells :, 2:]

    @property
    def data(self):
        dtype = type(self.convertData("0"))
        if dtype is str:
            dtype = object
        data = np.empty(self.dataCells.shape, dtype=dtype)
        for (row, column), cell in np.ndenumerate(self.dataCells):
            data[row, column] = self.convertData(cell.get())
        if self.maskBlanks:
            data = ma.masked_invalid(data)
        if dtype is object:
            data = data.astype(str)
        return data

    @property
    def initialGuesses(self):
        if not self.allowGuesses:
            raise RuntimeError(
                "initialGuesses() called on a Table instance that does not allow initial guesses"
            )

        dtype = type(self.convertData("0"))
        if dtype is str:
            dtype = object
        data = ma.empty(self.dataCells.shape, dtype=dtype)
        for (row, column), cell in np.ndenumerate(self.dataCells):
            value = cell.get()
            if value.startswith("~"):
                data[row, column] = self.convertData(value[1:])
            else:
                data[row, column] = ma.masked

        if dtype is object:
            data = data.astype(str)
        return data

    @data.setter
    def data(self, data):
        if self.dataCells.shape != data.shape:
            raise ValueError(
                "Requested data shape {data.shape} does not match table shape"
                " {self.dataCells.shape}."
            )
        for cell, datum in zip(self.dataCells.flat, data.flat):
            cell.set(datum)

    @property
    def rowTitles(self):
        return np.array([title.get() for title in self.cells[self.headerCells :, 1]])

    @rowTitles.setter
    def rowTitles(self, titles):
        for cell, title in zip(self.cells[self.headerCells :, 1], titles):
            cell.set(title)

    @property
    def columnTitles(self):
        return np.array([title.get() for title in self.cells[1, 2:]])

    @columnTitles.setter
    def columnTitles(self, titles):
        # make sure the number of columns matches the number of titles
        difference = len(titles) - (self.cells.shape[1] - 2)
        if difference > 0:
            for _ in range(difference):
                self.addColumn()
        elif difference < 0:
            for _ in range(difference):
                self.deleteColumn(self.cells.shape[1])
        if "titles" in self.columnOptions or "readonlyTitles" in self.columnOptions:
            for cell, title in zip(self.cells[1, 2:], titles):
                cell.set(title)


class ButtonFrame(ttk.Frame):
    def __init__(self, master, reset, save, cancel, *args, **kwargs):
        super().__init__(master, borderwidth=padding, *args, **kwargs)

        self.save = save

        self.resetButton = ttk.Button(
            self, text="Reset", command=reset, style="danger.TButton"
        )
        self.resetButton.pack(side="left", padx=padding)

        self.saveButton = ttk.Button(
            self, text="OK", command=self.trySave, style="success.TButton"
        )
        self.saveButton.pack(side="right", padx=padding)

        self.cancelButton = ttk.Button(
            self, text="Cancel", command=cancel, style="secondary.TButton"
        )
        self.cancelButton.pack(side="right", padx=padding)

    def trySave(self):
        try:
            self.save()
        except Exception as e:
            mb.showerror(title="Could not save data", message=e, parent=self)


class WrappedLabel(ttk.Frame):
    # A label that automatically wraps text to its allocated width, and only propagates
    # its height.
    def __init__(self, master, *args, **kwargs):
        super().__init__(master)
        self.pack_propagate(False)
        self.label = ttk.Label(self, *args, **kwargs)
        self.label.pack(expand=True, fill="both")
        self.label.bind("<Configure>", self.callback)

    def callback(self, event):
        self.label.configure(wraplength=event.width - 2 * self.labelPadding - 4)
        self.configure(height=self.label.winfo_reqheight())

    @property
    def labelPadding(self):
        # The Tk C API returns the padding as a "pixels" object, so need to convert it
        # to string first before converting to int.
        padding = self.label.cget("padding")
        if type(padding) is tuple:
            return int(str(padding[0]))
        else:
            # no padding set
            return 0



import numpy as np
from scipy.optimize import minimize
from scipy.signal import find_peaks

titrationAttributes = (
    "title",
    "rawData",
    "continuousRange",
    "hasSignalTitles",
    "hasAdditionTitles",
    "additionTitles",
    "signalTitles",
    "transposeData",
    "yQuantity",
    "yUnit",
    "xQuantity",
    "xUnit",
    "fitResult",
    "lastKVars",
    "lastTotalConcVars",
    "lastKs",
    "lastTotalConcs",
    "lastSpeciesConcs",
    "lastSignalVars",
    "lastFittedSpectra",
    "lastFittedCurves",
)


class Titration:
    def __init__(self, title="Titration"):
        self.title = title
        self.continuousRange = np.array([-np.inf, np.inf])
        self.continuous = False
        self.hasSignalTitles = False
        self.hasAdditionTitles = False
        self.transposeData = False

    @property
    def numAdditions(self):
        return self.processedData.shape[0]

    @property
    def numSignals(self):
        return self.processedData.shape[1]

    @property
    def columnFilter(self):
        if self.continuous:
            from_, to = self.continuousRange
            return (self.signalTitles >= from_) & (self.signalTitles <= to)

        else:
            return slice(None)

    @property
    def processedData(self):
        return self.rawData[:, self.columnFilter]

    @property
    def processedSignalTitles(self):
        if self.hasSignalTitles:
            return self.signalTitles[self.columnFilter]
        else:
            return np.array(
                ["Signal " + str(i + 1) for i in range(self.processedData.shape[1])]
            )

    @property
    def processedSignalCount(self):
        return self.processedData.shape[1]

    @property
    def signalTitles(self):
        if self.hasSignalTitles:
            return self._signalTitles
        else:
            return np.array(
                ["Signal " + str(i + 1) for i in range(self.rawData.shape[1])]
            )

    @signalTitles.setter
    def signalTitles(self, titles):
        # TODO: replace flag with just setting titles to None
        if titles.size == 0:
            self.hasSignalTitles = False
            self.continuous = False
            return
        self.hasSignalTitles = True
        try:
            _signalTitles = np.array(titles, dtype=float)
        except ValueError:
            self._signalTitles = np.array(titles, dtype=str)
            self.continuous = False
            return

        self.continuous = True

        averageStep = abs(
            (_signalTitles[-1] - _signalTitles[0]) / (len(_signalTitles) - 1)
        )
        # Round to 3 significant figures
        self.averageStep = np.round(
            averageStep, int(np.ceil(-np.log10(averageStep)) + 2)
        )
        if self.averageStep >= 1.00:
            self.signalTitlesDecimals = 0
        else:
            leadingZeros = int(np.ceil(-np.log10(self.averageStep)))
            self.signalTitlesDecimals = (
                leadingZeros
                - 1
                + len(
                    str(round(self.averageStep * 10 ** (leadingZeros + 2))).rstrip("0")
                )
            )
        self._signalTitles = np.round(_signalTitles, self.signalTitlesDecimals)

    @property
    def signalTitlesStrings(self):
        if self.signalTitles.dtype == float:
            return np.array(
                [
                    f"{title:.{self.signalTitlesDecimals}f}"
                    for title in self.signalTitles
                ]
            )
        else:
            return self.signalTitles.astype(str)

    @property
    def processedSignalTitlesStrings(self):
        if self.processedSignalTitles.dtype == float:
            return np.array(
                [
                    f"{title:.{self.signalTitlesDecimals}f}"
                    for title in self.processedSignalTitles
                ]
            )
        else:
            return self.processedSignalTitles.astype(str)

    @property
    def additionTitles(self):
        if self.hasAdditionTitles:
            return self._additionTitles
        else:
            return np.array(
                ["Addition " + str(i + 1) for i in range(self.rawData.shape[0])]
            )

    @additionTitles.setter
    def additionTitles(self, additionTitles):
        if additionTitles.size == 0:
            return
        self.hasAdditionTitles = True
        self._additionTitles = additionTitles

    def getPeakIndices(self, maxPeaks=4, maxShoulderPeaks=2, threshold=0.1):
        numSignals = self.processedData.shape[1]
        if numSignals <= maxPeaks + maxShoulderPeaks:
            return np.arange(numSignals)

        # get the total movement for each signal
        movement = abs(np.diff(self.processedData, axis=0)).sum(axis=0)
        # get the range for each signal
        totalRange = np.abs(
            np.max(self.processedData, axis=0) - np.min(self.processedData, axis=0)
        )
        # find the signals with the largest total movement
        peakIndices, peakProperties = find_peaks(
            np.pad(movement, [1, 1], mode="minimum"), prominence=0
        )
        peakIndices -= 1
        prominences = peakProperties["prominences"]

        # select the four most prominent peaks
        largestFilter = prominences.argsort()[-maxPeaks:]
        largestPeakIndices = peakIndices[largestFilter]

        # Shoulder peaks can appear as inflection points rather than maxima.
        # We'll add the two most prominent inflection points from a first-order
        # approximation of the first derivative of the total movement:
        leftShoulders, leftShoulderProperties = find_peaks(
            -np.gradient(movement), height=[None, 0], prominence=0
        )
        rightShoulders, rightShoulderProperties = find_peaks(
            np.gradient(movement), height=[None, 0], prominence=0
        )

        # Filter out duplicates, which happens when the peak height is exactly 0
        filter = np.isin(rightShoulders, leftShoulders, invert=True)
        shoulderIndices = np.concatenate([leftShoulders, rightShoulders[filter]])
        shoulderProminences = np.concatenate(
            [
                leftShoulderProperties["prominences"],
                rightShoulderProperties["prominences"][filter],
            ]
        )

        # filter out shoulder peaks that are too close to the main peaks
        filter = [
            index not in peakIndices
            and index + 1 not in peakIndices
            and index - 1 not in peakIndices
            for index in shoulderIndices
        ]
        shoulderIndices = shoulderIndices[filter]
        shoulderProminences = shoulderProminences[filter]

        # select the two most prominent inflection points
        filter = shoulderProminences.argsort()[-maxShoulderPeaks:]
        largestShoulderIndices = shoulderIndices[filter]

        # combine the two arrays, without duplicates, and sort them
        largestPeakIndices = np.sort(
            np.unique(np.concatenate((largestPeakIndices, largestShoulderIndices)))
        )
        if len(largestPeakIndices) == 0:
            return np.array([np.argmax(movement)])

        # discard peaks that don't move far enough compared to the other peaks
        peaksRange = totalRange[largestPeakIndices]
        return largestPeakIndices[peaksRange >= np.max(peaksRange) * threshold]

    def optimisationFunc(self, ksAndTotalConcs):
        # scipy.optimize optimizes everything as a single array, so split it
        kVars = ksAndTotalConcs[: self.equilibriumConstants.variableCount]
        self.lastKVars = kVars
        totalConcVars = ksAndTotalConcs[self.equilibriumConstants.variableCount :]
        self.lastTotalConcVars = totalConcVars

        # get all Ks and total concs, as some are fixed and thus aren't passed
        # to the function as arguments
        speciationVars = self.equilibriumConstants.run(kVars)
        self.lastKs = speciationVars
        totalConcs = self.totalConcentrations.run(totalConcVars)
        self.lastTotalConcs = totalConcs

        speciesConcs = self.speciation.run(speciationVars, totalConcs)
        self.lastSpeciesConcs = speciesConcs

        contributingSpeciesFilter = self.contributingSpecies.run()
        signalVars, contributorsCountPerMolecule = self.contributors.run(speciesConcs)
        self.lastSignalVars = signalVars

        proportionalSignalVars = self.proportionality.run(
            signalVars, contributorsCountPerMolecule
        )

        knownSpectra = self.knownSignals.run()

        self.lastFittedSpectra, residuals, self.lastFittedCurves = self.fitSignals.run(
            proportionalSignalVars, knownSpectra
        )

        combinedResiduals = np.sqrt(np.sum(residuals))
        self.lastResiduals = combinedResiduals

        return combinedResiduals

    def optimisationFuncLog(self, logKsAndTotalConcs):
        ksAndTotalConcs = 10**logKsAndTotalConcs
        return self.optimisationFunc(ksAndTotalConcs)

    def optimise(self, callback=None):
        initialGuessKs = np.log10(self.equilibriumConstants.variableInitialGuesses)
        initialGuessConcs = np.log10(self.totalConcentrations.variableInitialGuesses)
        initialGuess = np.concatenate((initialGuessKs, initialGuessConcs))

        result = minimize(
            self.optimisationFuncLog,
            x0=initialGuess,
            method="nelder-mead",
            callback=callback,
        )
        # to make sure the last fit is the optimal one
        self.optimisationFuncLog(result.x)
        return result.x

    def fitData(self, callback=None):
        self.fitResult = 10 ** self.optimise(callback)



from tkinter import font

padding = 5
cellWidth = 17

figureParams = {"scale": 100, "x": 640, "y": 480}

defaultFigureParams = {"scale": 100, "x": 640, "y": 480}


def __getattr__(name):
    if name in ["boldFont", "italicFont"]:
        initFonts()
        return globals()[name]
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


def initFonts():
    global boldFont, italicFont

    boldFont = font.nametofont("TkTextFont").copy()
    boldFont["weight"] = "bold"

    italicFont = font.nametofont("TkTextFont").copy()
    italicFont["slant"] = "italic"



from abc import abstractmethod

import numpy as np
from numpy import ma

from . import moduleFrame


class Proportionality(moduleFrame.Strategy):
    requiredAttributes = ()

    @abstractmethod
    def run(self, contributorConcs, contributorsCountPerMolecule):
        pass


class GetConcs(Proportionality):
    def run(self, contributorConcs, contributorsCountPerMolecule):
        return contributorConcs


class GetFraction(Proportionality):
    def run(self, contributorConcs, contributorsCountPerMolecule):
        splitIndices = np.cumsum(contributorsCountPerMolecule, axis=-1)[:-1]
        concsPerMolecule = np.hsplit(contributorConcs, splitIndices)
        proportionalConcs = np.hstack(
            [
                concs / np.sum(concs, axis=-1, keepdims=True)
                for concs in concsPerMolecule
            ]
        )
        return ma.masked_invalid(proportionalConcs)


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Experimental Data"
    dropdownLabelText = "What are the signals proportional to?"
    dropdownOptions = {
        "Concentration (slow exchange)": GetConcs,
        "Mole fraction (fast exchange)": GetFraction,
    }
    attributeName = "proportionality"
    setDefault = False



import ctypes
import importlib.resources as res
import tkinter as tk
import tkinter.filedialog as fd
import tkinter.messagebox as mb
import tkinter.ttk as ttk
from tkinter import font

import ttkbootstrap

from . import __version__
from . import windowsHighDpiPatch
from .style import padding

try:
    appId = "daniilS.musketeer"
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(appId)
except Exception:
    pass


fullErrorMessages = True
olderror = mb.showerror


# TODO: convert to simple error message, with an option to expand to the full one
def newshowerror(title, message, *args, **kwargs):
    import traceback

    if fullErrorMessages and isinstance(message, Exception):
        message = traceback.format_exc()
    return olderror(title, message, *args, **kwargs)


mb.showerror = newshowerror


# need to keep a reference to the Style object so that image-based widgets
# appear correctly
ttkStyle = ttkbootstrap.Style(theme="lumen")
root = ttkStyle.master
root.geometry("1000x600")

windowsHighDpiPatch.setEnhancedDpiScaling(root)

try:
    # Windows and OS X
    root.state("zoomed")
except tk.TclError:
    # X11
    root.attributes("-zoomed", True)
root.title(f"Musketeer {__version__}")


# notes on preferred icon size on Windows:
# title bar: prefers 16px, actual size 16px (or 24px at 1.5x scaling)
# if the icon set without calling update() first, the title bar also prefers 32, but
# will be blurry
# task bar: prefers 32px, actual size 24px (or 36px at 1.5x scaling)
# if the preferred size is not provided, the first file provided will be used
# the least common multiple of 16, 24, and 36 is 144, but using that causes the title
# bar icon to be blurry
# 48px seems to give the best quality at all scaling factors

# on Linux, tk documentation recommends providing no more than 2 icons, placing the
# largest one first
# on MacOS, only the first icon is used, and it is recommended to provide as large an
# icon as possible

try:
    try:
        iconData48 = (res.files(__package__) / "logo 48px.png").read_bytes()
        iconData512 = (res.files(__package__) / "logo 512px.png").read_bytes()
    except AttributeError:
        # Python 3.8 compatibility
        iconData48 = res.read_binary(__package__, "logo 48px.png")
        iconData512 = res.read_binary(__package__, "logo 512px.png")
    icon48 = tk.PhotoImage(data=iconData48)
    icon512 = tk.PhotoImage(data=iconData512)

    if root._windowingsystem == "win32":
        # Call before update to make sure the icon appears straight away (otherwise the
        # default Tk icon may appear for a few seconds first). Call after update is
        # necessary to make the title bar icon not blurry.
        root.iconphoto(True, icon48, icon512)
        root.update()
        root.iconphoto(True, icon48, icon512)
    else:
        root.iconphoto(True, icon512, icon48)
except Exception:
    # Should currently never happen, but if anything changes about the API in
    # the future, we can survive without an icon.
    pass
frame = ttk.Frame(root, padding=padding)
frame.pack(expand=True, fill="both")


# Especially on the first startup, when matplotlib needs to build the font cache,
# importing everything may take a while. So we defer all other imports until now.
root.tk.eval("tk busy .")
root.update()

from pathlib import PurePath

import numpy as np
from ttkbootstrap.widgets import InteractiveNotebook

from . import patchMatplotlib
from .style import defaultFigureParams, figureParams
from .table import ButtonFrame
from .titration import Titration
from .titrationFrame import TitrationFrame

patchMatplotlib.applyPatch()


class DpiPopup(tk.Toplevel):
    def __init__(self, master, saveCallback, *args, **kwargs):
        super().__init__(
            master,
            padx=padding,
            pady=padding,
            *args,
            **kwargs,
        )
        self.title("Change figure DPI")
        self.resizable(False, False)
        self.saveCallback = saveCallback

        scaleLabel = ttk.Label(self, text="Scale:", justify="left", state="normal")
        scaleLabel.grid(row=0, column=0, sticky="ew")

        self.scaleDropdown = ttk.Combobox(
            self,
            values=[f"{i}%" for i in [60, 80, 100, 125, 150, 200]],
            justify="right",
            width=5,
        )
        self.scaleDropdown.grid(row=0, column=2, sticky="ew")

        resLabel = ttk.Label(self, text="Resolution:", justify="left")
        resLabel.grid(row=1, column=0, sticky="ew")

        self.resDropdown = ttk.Combobox(
            self,
            values=[
                f"{int(y*4/3)} x {y}" for y in [240, 360, 480, 600, 960, 1200, 1440]
            ],
            justify="right",
            width=10,
            state="readonly",
        )
        self.resDropdown.grid(row=1, column=2, sticky="ew")

        self.rowconfigure(0, pad=10)
        self.rowconfigure(1, pad=10)

        self.columnconfigure(0, pad=10)
        self.columnconfigure(1, weight=1)

        self.initValues()

        buttonFrame = ButtonFrame(self, self.reset, self.save, self.destroy)
        buttonFrame.saveButton.configure(text="Save")

        buttonFrame.applyButton = ttk.Button(
            buttonFrame, text="Apply", command=self.apply, style="success.TButton"
        )
        buttonFrame.applyButton.pack(
            side="right", after=buttonFrame.saveButton, padx=padding
        )

        buttonFrame.grid(row=2, column=0, columnspan=3, sticky="esw")

    def reset(self):
        self.scaleDropdown.set(f"{defaultFigureParams['scale']}%")
        self.resDropdown.set(f"{defaultFigureParams['x']} x {defaultFigureParams['y']}")

    def initValues(self):
        self.scaleDropdown.set(f"{figureParams['scale']}%")
        self.resDropdown.set(f"{figureParams['x']} x {figureParams['y']}")

    def save(self):
        if self.apply():  # returns False if failed to apply
            self.destroy()

    def apply(self):
        try:
            scale = int(self.scaleDropdown.get().strip("%"))
            x, y = [int(i) for i in self.resDropdown.get().split(" x ")]
            figureParams.update(scale=scale, x=x, y=y)
            self.saveCallback()
            return True
        except Exception as e:
            mb.showerror(title="Could not apply settings", message=e, parent=self)
            return False


class TitrationsNotebook(InteractiveNotebook):
    def __init__(self, master, *args, **kwargs):
        super().__init__(
            master,
            padding=padding,
            newtab=self.newFile,
            style="Flat.Interactive.TNotebook",
            *args,
            **kwargs,
        )
        self.createMenuBar()
        self.createQuickStartMenu()
        self.select(str(self._newtab_frame))

    def forget(self, index, *args, **kwargs):
        # Manually change to a different tab before a tab is closed, to prevent the
        # quickstart menu from appearing briefly right after a tab is closed but before
        # the next tab has been selected
        next_tab = index + 1
        if (
            self._has_newtab_button and next_tab == self._last_tab
        ) or index == self._last_tab:
            next_tab = index - 1
        if 0 <= next_tab <= self._last_tab:
            self.select(next_tab)
        return super().forget(index, *args, **kwargs)

    def createQuickStartMenu(self):
        self.quickStartFrame = ttk.Frame(self)
        self.quickStartFrame.pack(expand=True, fill="none")

        self.quickStartFont = font.nametofont("TkTextFont").copy()
        self.quickStartFont["size"] = round(self.quickStartFont["size"] * 1.33)
        ttkStyle.configure("large.success.TButton", font=self.quickStartFont)
        ttkStyle.configure("large.TButton", font=self.quickStartFont)

        self.newFileButton = ttk.Button(
            self.quickStartFrame,
            text="Create a new fit file",
            command=self.newFile,
            style="large.success.TButton",
        )
        self.newFileButton.pack(fill="x")
        self.orLabel = ttk.Label(
            self.quickStartFrame, text="or", font=self.quickStartFont
        )
        self.orLabel.pack(pady=5)
        self.OpenFileButton = ttk.Button(
            self.quickStartFrame,
            text="Open an existing file",
            command=self.openFile,
            style="large.TButton",
        )
        self.OpenFileButton.pack(fill="x")

    def createMenuBar(self):
        self.menuBar = tk.Menu(self, tearoff=False)

        fileMenu = tk.Menu(self.menuBar, tearoff=False)
        self.menuBar.add_cascade(label="File", menu=fileMenu, underline=0)

        self.addMenuCommand(fileMenu, "New", self.newFile)
        self.addMenuCommand(fileMenu, "Open", self.openFile)
        self.addMenuCommand(fileMenu, "Save", self.saveFile)
        self.addMenuCommand(
            fileMenu, "Save As", self.saveFileAs, keys=("Shift", "s"), underline=5
        )

        editMenu = tk.Menu(self.menuBar, tearoff=False)
        self.menuBar.add_cascade(label="Edit", menu=editMenu, underline=0)
        editMenu.add_command(
            label="Change figure DPI", command=self.editDpi, underline=0
        )

        self.winfo_toplevel().config(menu=self.menuBar)

    def addMenuCommand(self, fileMenu, label, command, keys=None, underline=0):
        if self._windowingsystem == "aqua":
            accelerator = "Command"
            key = accelerator
        else:
            accelerator = "Ctrl"
            key = "Control"

        if keys is None:
            keys = (label[0].lower(),)

        fileMenu.add_command(
            label=label,
            command=command,
            accelerator=f"{accelerator}+{'+'.join(keys).title()}",
            underline=underline,
        )
        self.winfo_toplevel().bind(f"<{key}-{'-'.join(keys)}>", command)

    def newFile(self, *args):
        titration = Titration("Fit 1")
        titration.rawData = np.empty((0, 0))

        titrationFrame = TitrationFrame(self, padding=padding)
        self.add(titrationFrame, text="New Titration", sticky="nesw")
        self.select(str(titrationFrame))
        titrationFrame.loadTitration(titration)

    def saveFile(self, *args):
        self.nametowidget(self.select()).saveFile()

    def saveFileAs(self, *args):
        self.nametowidget(self.select()).saveFile(saveAs=True)

    def openFile(self, *args):
        filePath = fd.askopenfilename(
            title="Open file",
            filetypes=[("Musketeer files", "*.fit"), ("All files", "*.*")],
        )
        if filePath == "":
            return

        try:
            titration = np.load(filePath, allow_pickle=False)
        except Exception as e:
            mb.showerror(title="Failed to read file", message=e, parent=self)
            return

        self.tk.eval("tk busy .")
        self.update()

        titrationFrame = TitrationFrame(self, filePath, padding=padding)
        self.add(titrationFrame, text=PurePath(filePath).name, sticky="nesw")
        self.select(str(titrationFrame))
        titrationFrame.loadTitration(titration)

        self.tk.eval("tk busy forget .")

    def editDpi(self, *args):
        popup = DpiPopup(self, self.updateDpi)
        popup.withdraw()
        self.update()
        root = self.winfo_toplevel()
        x = root.winfo_x() + root.winfo_width() / 2 - popup.winfo_width() / 2
        y = root.winfo_y() + root.winfo_height() / 2 - popup.winfo_height() / 2
        popup.geometry(f"+{int(x)}+{int(y)}")
        popup.transient(self)
        popup.grab_set()
        popup.deiconify()
        popup.wait_window()

    def updateDpi(self):
        for tab in self.tabs():
            if isinstance(titrationFrame := self.nametowidget(tab), TitrationFrame):
                titrationFrame.updateDpi()
        self.update()


notebook = TitrationsNotebook(frame)
notebook.pack(expand=True, fill="both")

root.tk.eval("tk busy forget .")
root.mainloop()



import sys

import musketeer
from cx_Freeze import Executable, setup

build_options = {
    "packages": [],
    "excludes": [],
    "include_files": [] if sys.platform == "win32" else ["logo 512px.png"],
}
msi_options = {
    "install_icon": "logo 48px.ico",
    "upgrade_code": "{C327C15B-6058-313C-AADF-E979233602A5}",
    "summary_data": {"author": "Daniil Soloviev"},
}


base = "Win32GUI" if sys.platform == "win32" else None

executables = [
    Executable(
        "musketeer_loader.py",
        base=base,
        target_name="Musketeer",
        shortcut_name="Musketeer",
        shortcut_dir="ProgramMenuFolder",
        icon="logo 48px.ico" if sys.platform == "win32" else "logo 512px.png",
    )
]

setup(
    name="Musketeer",
    version=musketeer.__version__,
    description="Musketeer",
    options={"build_exe": build_options, "bdist_msi": msi_options},
    executables=executables,
)



import tkinter as tk
import tkinter.filedialog as fd
import tkinter.messagebox as mb
import tkinter.ttk as ttk
import warnings
from collections import namedtuple

import numpy as np
from numpy import ma
from tksheet import Sheet

from . import moduleFrame
from . import titrationReader
from .style import padding
from .table import ButtonFrame

Params = namedtuple(
    "Params",
    ("yQuantity", "yUnit", "xQuantity", "xUnit"),
    defaults=[""] * 4,
)

predefinedParams = {
    "UV-Vis": Params("Abs", "AU", "λ", "nm"),
    "NMR": Params("δ", "ppm"),
}


class EditDataPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter/edit spectroscopic data")

        frame = ttk.Frame(self, padding=15)
        frame.pack(expand=True, fill="both")

        paramsFrame = ttk.Frame(frame)
        paramsFrame.pack(expand=False, fill="both")

        paramsFrameLeft = ttk.Frame(paramsFrame)
        paramsFrameLeft.grid(row=0, column=0, sticky="ne", padx=5, pady=5)
        paramsFrame.grid_columnconfigure(0, weight=1)

        self.hasAdditionTitles = tk.BooleanVar()
        self.additionTitlesCheckbutton = ttk.Checkbutton(
            paramsFrameLeft, text="Addition titles", variable=self.hasAdditionTitles
        )
        self.hasSignalTitles = tk.BooleanVar()
        self.signalTitlesCheckbutton = ttk.Checkbutton(
            paramsFrameLeft, text="Signal titles", variable=self.hasSignalTitles
        )

        self.additionsRowsRadiobutton = ttk.Radiobutton(
            paramsFrameLeft, value=0, text="Rows are additions, columns are signals"
        )
        self.additionsColumnsRadiobutton = ttk.Radiobutton(
            paramsFrameLeft, value=1, text="Rows are signals, columns are additions"
        )

        self.additionTitlesCheckbutton.pack(pady=2.5, fill="x")
        self.signalTitlesCheckbutton.pack(pady=2.5, fill="x")
        self.additionsRowsRadiobutton.pack(pady=2.5, fill="x")
        self.additionsColumnsRadiobutton.pack(pady=2.5, fill="x")

        paramsFrameRight = ttk.Frame(paramsFrame)
        paramsFrameRight.grid(row=0, column=1, sticky="nw", padx=5, pady=5)
        paramsFrame.grid_columnconfigure(1, weight=1)

        optionMenuLabel = ttk.Label(paramsFrameRight, text="Autofill quantities/units:")
        optionMenuLabel.grid(row=0, column=0, sticky="w")

        self.optionMenuVar = tk.StringVar(self)
        optionMenu = ttk.OptionMenu(
            paramsFrameRight,
            self.optionMenuVar,
            None,
            *predefinedParams.keys(),
            style="Outline.TMenubutton",
            command=self.setParams,
        )
        optionMenu.configure(width=max([len(s) for s in predefinedParams]) + 1)
        optionMenu.grid(row=0, column=1, pady=2.5)

        self.yQuantityLabel = ttk.Label(paramsFrameRight, text="Measured quantity:")
        self.yQuantityLabel.grid(row=1, column=0, sticky="w")
        self.yUnitLabel = ttk.Label(paramsFrameRight, text="Unit:")
        self.yUnitLabel.grid(row=1, column=1, sticky="w")
        self.yQuantity = tk.StringVar(self)
        self.yUnit = tk.StringVar(self)
        self.yQuantityWidget = ttk.Entry(paramsFrameRight, textvariable=self.yQuantity)
        self.yQuantityWidget.grid(row=2, column=0, sticky="w")
        self.yUnitWidget = ttk.Entry(
            paramsFrameRight, width=10, textvariable=self.yUnit
        )
        self.yUnitWidget.grid(row=2, column=1, sticky="w")

        self.xQuantityLabel = ttk.Label(
            paramsFrameRight, text="Continuous signals x-axis quantity:"
        )
        self.xQuantityLabel.grid(row=3, column=0, sticky="w")
        self.xUnitLabel = ttk.Label(paramsFrameRight, text="Unit:")
        self.xUnitLabel.grid(row=3, column=1, sticky="w")
        self.xQuantity = tk.StringVar(self)
        self.xUnit = tk.StringVar(self)
        self.xQuantityWidget = ttk.Entry(paramsFrameRight, textvariable=self.xQuantity)
        self.xQuantityWidget.grid(row=4, column=0, sticky="w")
        self.xUnitWidget = ttk.Entry(
            paramsFrameRight, width=10, textvariable=self.xUnit
        )
        self.xUnitWidget.grid(row=4, column=1, sticky="w")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.sheet = Sheet(
                frame,
                data=[[]],
                expand_sheet_if_paste_too_big=True,
            )

        self.sheet.pack(side="top", expand=True, fill="both")
        self.sheet.enable_bindings()

        buttonFrame = ButtonFrame(self, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        loadButton = ttk.Button(
            buttonFrame, text="Import from file", command=self.loadCSV
        )
        loadButton.pack(side="left", padx=padding)

        self.populateDefault()

    def populateDefault(self):
        titration = self.titration
        self.hasAdditionTitles.set(titration.hasAdditionTitles)
        self.hasSignalTitles.set(titration.hasSignalTitles)

        if not self.titration.transposeData:
            self.additionsRowsRadiobutton.invoke()
        else:
            self.additionsColumnsRadiobutton.invoke()

        self.optionMenuVar.set("")

        for param in Params._fields:
            if hasattr(titration, param):
                getattr(self, param).set(getattr(titration, param))
            else:
                getattr(self, param).set("")

        data = titration.rawData.copy()
        if data.shape == (0, 0):
            data = np.full((16, 5), "")
        else:
            formattedData = data.astype(str)
            formattedData[np.isnan(data)] = ""
            data = formattedData
        if titration.hasAdditionTitles and titration.hasSignalTitles:
            data = np.c_[titration.additionTitles, data]
            data = np.r_[[np.insert(titration.signalTitlesStrings, 0, "")], data]
        elif titration.hasAdditionTitles:
            data = np.c_[titration.additionTitles, data]
        elif titration.hasSignalTitles:
            data = np.r_[[titration.signalTitlesStrings], data]

        if titration.transposeData:
            data = data.T
        self.sheet.set_sheet_data(data.tolist())

    def setParams(self, selection):
        params = predefinedParams[selection]
        for k, v in params._asdict().items():
            if v is not None:
                getattr(self, k).set(v)

    def reset(self):
        self.populateDefault()

    def processData(self):
        titration = self.titration
        data = np.array(self.sheet.get_sheet_data())
        data = data[~np.all(data == "", 1), :][:, ~np.all(data == "", 0)]

        titration.hasSignalTitles = self.signalTitlesCheckbutton.instate(["selected"])
        if not titration.hasSignalTitles:
            titration.continuous = False
        titration.hasAdditionTitles = self.additionTitlesCheckbutton.instate(
            ["selected"]
        )
        titration.transposeData = self.additionsColumnsRadiobutton.instate(["selected"])

        if titration.transposeData:
            data = data.T

        if titration.hasAdditionTitles and titration.hasSignalTitles:
            titration.additionTitles = data[1:, 0]
            titration.signalTitles = data[0, 1:]
            rawData = data[1:, 1:]
        elif titration.hasAdditionTitles:
            titration.additionTitles = data[:, 0]
            rawData = data[:, 1:]
        elif titration.hasSignalTitles:
            titration.signalTitles = data[0, :]
            rawData = data[1:, :]
        else:
            rawData = data

        rawData = rawData.astype(object)  # to allow setting values to "nan"
        rawData[rawData == ""] = "nan"
        return ma.masked_invalid(rawData.astype(float))

    def saveData(self):
        self.titration.rawData = self.processData()
        for param in Params._fields:
            setattr(self.titration, param, getattr(self, param).get())

        self.saved = True
        self.destroy()

    def loadCSV(self):
        fileType = tk.StringVar(self)
        fileReaders = titrationReader.fileReaders
        filePath = fd.askopenfilename(
            title="Import spectroscopic data from file",
            filetypes=fileReaders[:, :-1].tolist(),
            typevariable=fileType,
        )
        if filePath == "":
            return
        fileReader = fileReaders[fileReaders[:, 0] == fileType.get(), -1].item()
        try:
            data, additionTitles, signalTitles, defaultParams = fileReader(
                filePath, self
            )
        except Exception as e:
            mb.showerror(title="Failed to read file", message=e, parent=self)
            return

        if (additionTitles is not None) and (signalTitles is not None):
            data = np.c_[additionTitles, data]
            data = np.r_[[np.insert(signalTitles, 0, "")], data]
            self.hasAdditionTitles.set(True)
            self.hasSignalTitles.set(True)
        elif additionTitles is not None:
            data = np.c_[additionTitles, data]
            self.hasAdditionTitles.set(True)
            self.hasSignalTitles.set(False)
        elif signalTitles is not None:
            data = np.r_[[signalTitles], data]
            self.hasAdditionTitles.set(False)
            self.hasSignalTitles.set(True)
        else:
            self.hasAdditionTitles.set(False)
            self.hasSignalTitles.set(False)

        if defaultParams is not None:
            self.optionMenuVar.set(defaultParams)
            self.setParams(defaultParams)

        self.additionsRowsRadiobutton.invoke()

        self.sheet.set_sheet_data(data.tolist())



import math
import tkinter.ttk as ttk

import numpy as np

from . import moduleFrame
from .scrolledFrame import ScrolledFrame
from .style import padding
from .table import ButtonFrame, Table, WrappedLabel


class Contributors:
    requiredAttributes = ()


class ContributorConcs(moduleFrame.Strategy):
    requiredAttributes = (
        "contributorsMatrix",
        "outputNames",
        "contributorsCountPerMolecule",
    )

    def run(self, speciesConcs):
        return (
            speciesConcs @ self.contributorsMatrix.T,
            self.contributorsCountPerMolecule,
        )


class ContributorsTable(Table):
    def __init__(
        self, master, outputNames, speciesNames, contributorsMatrix, speciesFilter
    ):
        if len(speciesNames) != len(speciesFilter):
            raise ValueError(
                f"Lengths of speciesNames ({len(speciesNames)}) and speciesFilter"
                f" ({len(speciesFilter)}) do not match."
            )

        self.speciesFilter = speciesFilter
        self.width = max(
            [len(name) for name in np.concatenate([outputNames, speciesNames])]
        )

        if contributorsMatrix.shape[1] == len(speciesFilter):
            data = contributorsMatrix[:, speciesFilter]
            rowTitles = outputNames
        else:
            data = np.eye(np.count_nonzero(speciesFilter), dtype=int)
            rowTitles = speciesNames[speciesFilter]
        columnTitles = speciesNames[speciesFilter]

        super().__init__(
            master,
            0,
            0,
            columnTitles,
            rowOptions=("titles", "new", "delete"),
            columnOptions=("readonlyTitles",),
            boldTitles=True,
        )
        for name, contributions in zip(rowTitles, data):
            self.addRow(name, contributions)

    def newRow(self):
        defaultEntries = np.full(self.dataCells.shape[1], "0")
        self.addRow("New state", defaultEntries)

    def convertData(self, number):
        # if all entries happen to be integers, this is handled by processData()
        return float(number)

    def processData(self):
        matrix = np.zeros([self.data.shape[0], len(self.speciesFilter)])
        matrix[:, self.speciesFilter] = self.data
        if np.all(matrix % 1 == 0):
            matrix = matrix.astype(int)

        contributorsMatrix = matrix
        outputNames = self.rowTitles
        # Returns an array so that ContributorConcsPopup can save it directly as a
        # popup attribute. ContributorConcsPerMoleculePopup will convert it to a long
        # array.
        contributorsCountPerMolecule = np.array([matrix.shape[0]])

        return contributorsMatrix, outputNames, contributorsCountPerMolecule


class ContributorConcsPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter the contributing states")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        contributorsLabel = WrappedLabel(
            self.frame,
            padding=padding * 2,
            text=(
                "On each row, enter a state that contributes to the observed"
                " signal. For each column, specify how many of the state that"
                " species contains."
            ),
        )
        contributorsLabel.pack(expand=False, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        scrolledFrame = ScrolledFrame(
            self.frame, max_width=self.winfo_toplevel().master.winfo_width() - 200
        )
        scrolledFrame.pack(expand=True, fill="both")
        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)
        self.createTable()

    def createTable(self):
        self.contributorsTable = ContributorsTable(
            self.innerFrame,
            self.titration.contributors.outputNames,
            self.titration.speciation.outputNames,
            self.titration.contributors.contributorsMatrix,
            self.titration.contributingSpecies.filter,
        )
        self.contributorsTable.pack(expand=True, fill="both")

    def reset(self):
        self.contributorsTable.destroy()
        self.createTable()

    def saveData(self):
        (
            self.contributorsMatrix,
            self.outputNames,
            self.contributorsCountPerMolecule,
        ) = self.contributorsTable.processData()

        self.saved = True
        self.destroy()


class ContributorConcsPerMoleculePopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter the contributing states")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        contributorsLabel = WrappedLabel(
            self.frame,
            padding=padding * 2,
            text=(
                "On each row, enter a state that contributes to the observed"
                " signal. For each column, specify how many of the state that"
                " species contains."
            ),
        )
        contributorsLabel.pack(expand=False, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        scrolledFrame = ScrolledFrame(
            self.frame, max_width=self.winfo_toplevel().master.winfo_width() - 200
        )
        scrolledFrame.pack(expand=True, fill="both")
        self.innerFrame = scrolledFrame.display_widget(ttk.Frame, stretch=True)
        self.createNotebook()

    def createNotebook(self):
        self.notebook = ttk.Notebook(self.innerFrame, style="Flat.TNotebook")
        self.notebook.pack(expand=True, fill="both", padx=padding, pady=padding)

        self.tables = []

        splitIndices = np.cumsum(
            self.titration.contributors.contributorsCountPerMolecule
        )[:-1]

        outputNamesPerMolecule = np.split(
            self.titration.contributors.outputNames, splitIndices
        )
        contributorsMatrixPerMolecule = np.vsplit(
            self.titration.contributors.contributorsMatrix, splitIndices
        )
        moleculeHasSignals = [
            index in self.titration.contributingSpecies.signalToMoleculeMap
            for index in range(self.titration.speciation.freeCount)
        ]

        for molecule, outputNames, contributorsMatrix, speciesFilter, hasSignals in zip(
            self.titration.speciation.freeNames,
            outputNamesPerMolecule,
            contributorsMatrixPerMolecule,
            self.titration.contributingSpecies.filter,  # already 1 row per molecule
            moleculeHasSignals,
        ):
            if hasSignals:
                table = ContributorsTable(
                    self.notebook,
                    outputNames,
                    self.titration.speciation.outputNames,
                    contributorsMatrix,
                    speciesFilter,
                )
                self.notebook.add(table, text=molecule)
                self.tables.append(table)
            else:
                self.tables.append(None)

    def reset(self):
        self.notebook.destroy()
        self.createNotebook()

    def saveData(self):
        contributorsMatrix = []
        outputNames = []
        contributorsCountPerMolecule = []
        for table in self.tables:
            if table is None:
                contributorsCountPerMolecule.append(np.array([0]))
            else:
                (
                    matrix,
                    names,
                    count,
                ) = table.processData()
                contributorsMatrix.append(matrix)
                outputNames.append(names)
                contributorsCountPerMolecule.append(count)

        self.contributorsMatrix = np.vstack(contributorsMatrix)
        self.outputNames = np.concatenate(outputNames)
        self.contributorsCountPerMolecule = np.concatenate(contributorsCountPerMolecule)

        self.saved = True
        self.destroy()


class ContributorConcsAll(ContributorConcs):
    @property
    def outputNames(self):
        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        filter = self.titration.contributingSpecies.filter
        if singleMoleculeIndex is None or type(singleMoleculeIndex) is int:
            return self.titration.speciation.outputNames[filter]
        elif type(singleMoleculeIndex) is np.ndarray:
            names = []
            for i in singleMoleculeIndex:
                moleculeName = self.titration.speciation.freeNames[i]
                allNames = np.array(
                    [
                        f"{moleculeName} in {complex}"
                        for complex in self.titration.speciation.outputNames
                    ]
                )
                allNames[i] = f"Free {moleculeName}"
                names.extend(allNames[filter[i]])
            return np.array(names)

    @property
    def contributorsMatrix(self):
        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        filter = self.titration.contributingSpecies.filter
        if singleMoleculeIndex is None:
            return np.diag(filter).astype(int)[filter]
        elif type(singleMoleculeIndex) is int:
            return self.matrixFromSingleMolecule(singleMoleculeIndex, filter)
        elif type(singleMoleculeIndex) is np.ndarray:
            # The index will currently always be equal to np.arange(freeCount), so a
            # matrix will be returned for each molecule. It may at some point become
            # more useful to instead only return matrices for the molecules that have
            # at least one signal mapped to them.
            return np.vstack(
                [
                    self.matrixFromSingleMolecule(i, filter[i])
                    for i in singleMoleculeIndex
                ]
            )

    @property
    def contributorsCountPerMolecule(self):
        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        filter = self.titration.contributingSpecies.filter
        if type(singleMoleculeIndex) is np.ndarray:
            return np.array(
                [
                    self.matrixFromSingleMolecule(i, filter[i]).shape[0]
                    for i in singleMoleculeIndex
                ]
            )
        else:
            return np.array([self.contributorsMatrix.shape[0]])

    def matrixFromSingleMolecule(self, index, filter):
        diagonal = abs(self.titration.speciation.outputStoichiometries[:, index])
        return np.diag(diagonal)[filter]


# TODO: fix for homodimers
class ContributorConcsIdentical(ContributorConcs):
    def calculateStates(self):
        moleculeCount = self.titration.speciation.freeCount
        freeStates = np.zeros(
            [self.titration.speciation.outputCount, moleculeCount], dtype=int
        )
        boundStates = np.zeros(
            [self.titration.speciation.outputCount, moleculeCount], dtype=int
        )

        # for each molecule, the maximum number of bonds it is known to form to each
        # other molecule
        # rows are the host molecules, columns are the guest molecules
        maximumValencyPerGuest = np.zeros([moleculeCount, moleculeCount], dtype=int)
        # maximum total number of bonds each molecule can form
        maximumTotalValency = np.zeros(moleculeCount, dtype=int)

        stoichiometries = self.titration.speciation.outputStoichiometries[
            self.titration.speciation.freeCount :
        ]
        # n-mer means host has n-1 binding sites for binding itself
        # 1 becomes 0, because that's the host molecule itself
        # -1 becomes -2 becomes 2, as polymers imply 2 binding sites
        guestStoichiometries = np.repeat(
            stoichiometries[np.newaxis, :, :], moleculeCount, axis=0
        )
        guestStoichiometries[range(moleculeCount), :, range(moleculeCount)] -= 1
        guestStoichiometries = np.abs(guestStoichiometries)

        # calculate maximum valency per guest and maximum total valency
        for host in range(moleculeCount):
            singleHost = np.abs(stoichiometries)[:, host] == 1
            maximumValencyPerGuest[host] = np.max(
                guestStoichiometries[host][singleHost, :], axis=0, initial=0
            )

            # We only consider guests that form binary complexes with the host
            formsBinaryComplex = np.zeros(moleculeCount, dtype=bool)
            for guest in range(moleculeCount):
                binaryComplex = np.zeros(moleculeCount, dtype=int)
                binaryComplex[host] += 1
                binaryComplex[guest] += 1
                formsBinaryComplex[guest] = np.any(
                    np.all(
                        # want to match -1 iff host == guest
                        np.abs(stoichiometries) == binaryComplex,
                        axis=1,
                    )
                )
            maximumValencyPerGuest[host, ~formsBinaryComplex] = 0

            # Find the maximum number of guests that can bind to one host molecule at
            # the same time, counting only those guests which can also form binary
            # complexes with the host.
            maximumTotalValency[host] = np.max(
                np.sum(
                    guestStoichiometries[host][singleHost, :][:, formsBinaryComplex],
                    axis=1,
                ),
                initial=0,
            )

        # calculate number of free and bound states for each molecule in each complex
        for host in range(moleculeCount):
            freeStates[host, host] = maximumTotalValency[host]
            for index, complex in enumerate(guestStoichiometries[host]):
                if stoichiometries[index, host] < 1:
                    # complexes handled separately below
                    continue
                else:
                    hostStoichiometry = stoichiometries[index, host]
                    maximumTotalValencyHost = (
                        maximumTotalValency[host] * hostStoichiometry
                    )
                valencyPerGuest = np.minimum(
                    complex * maximumValencyPerGuest[:, host],
                    maximumValencyPerGuest[host, :] * hostStoichiometry,
                )
                totalValency = np.minimum(
                    np.sum(valencyPerGuest),
                    maximumTotalValencyHost,
                )
                freeStates[moleculeCount + index, host] = (
                    maximumTotalValencyHost - totalValency
                )
                boundStates[moleculeCount + index, host] = totalValency

            complexIndices = np.where(stoichiometries[:, host] < 0)[0]
            terminal = complexIndices[0::2]
            internal = complexIndices[1::2]
            # Crude assumption for polymers with odd numbers of binding sites - users
            # are really expected to manually enter a contributors table in such cases.
            freeStates[moleculeCount + terminal, host] = math.ceil(
                maximumTotalValency[host] / 2
            )
            boundStates[moleculeCount + terminal, host] = math.floor(
                maximumTotalValency[host] / 2
            )
            freeStates[moleculeCount + internal, host] = 0
            boundStates[moleculeCount + internal, host] = maximumTotalValency[host]

        return freeStates, boundStates

    def getContributorsMatrixAndNames(self):
        freeStates, boundStates = self.calculateStates()
        allStates = np.empty(
            [
                2 * self.titration.speciation.freeCount,
                self.titration.speciation.outputCount,
            ],
            dtype=int,
        )
        allStates[0::2] = freeStates.T
        allStates[1::2] = boundStates.T

        allNames = np.concatenate(
            [
                [f"Free {molecule} site", f"Bound {molecule} site"]
                for molecule in self.titration.speciation.freeNames
            ]
        )

        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        if singleMoleculeIndex is None:
            relevantStates = allStates[:, self.titration.contributingSpecies.filter]
            rowFilter = np.any(relevantStates, axis=1)
        elif type(singleMoleculeIndex) is int:
            rowFilter = np.array([2 * singleMoleculeIndex, 2 * singleMoleculeIndex + 1])
        elif type(singleMoleculeIndex) is np.ndarray:
            rowFilter = np.empty(2 * len(singleMoleculeIndex), dtype=int)
            rowFilter[0::2] = 2 * singleMoleculeIndex
            rowFilter[1::2] = 2 * singleMoleculeIndex + 1
        return allStates[rowFilter], allNames[rowFilter]

    @property
    def outputNames(self):
        _, names = self.getContributorsMatrixAndNames()
        return names

    @property
    def contributorsMatrix(self):
        matrix, _ = self.getContributorsMatrixAndNames()
        return matrix

    @property
    def contributorsCountPerMolecule(self):
        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        if type(singleMoleculeIndex) is np.ndarray:
            return np.array([2] * len(singleMoleculeIndex))
        else:
            return np.array([self.contributorsMatrix.shape[0]])


class ContributorConcsCustom(ContributorConcs):
    @property
    def Popup(self):
        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        if type(singleMoleculeIndex) is np.ndarray:
            return ContributorConcsPerMoleculePopup
        else:
            return ContributorConcsPopup

    popupAttributes = (
        "contributorsMatrix",
        "outputNames",
        "contributorsCountPerMolecule",
    )


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Spectra"
    dropdownLabelText = "Specify relationship between fitted spectra?"
    dropdownOptions = {
        "All species have different spectra": ContributorConcsAll,
        "All binding sites have identical spectra": ContributorConcsIdentical,
        "Custom": ContributorConcsCustom,
    }
    attributeName = "contributors"



import csv
import tkinter as tk
import tkinter.filedialog as fd
import tkinter.ttk as ttk
import warnings

import numpy as np
from numpy import ma
from tksheet import Sheet

from . import moduleFrame
from .style import padding
from .table import ButtonFrame


class KnownSignals(moduleFrame.Strategy):
    requiredAttributes = ("knownSpectra",)

    def run(self):
        return self.knownSpectra


class KnownSpectraPopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter known spectra")

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.sheet = Sheet(
                self,
                empty_vertical=0,
                empty_horizontal=0,
                data=list(titration.knownSignals.knownSpectra.astype(str).filled("")),
                headers=list(titration.processedSignalTitlesStrings),
                row_index=list(titration.contributors.outputNames),
                set_all_heights_and_widths=True,
            )
        self.sheet.MT.configure(height=self.sheet.MT.row_positions[-1] + 1)
        self.sheet.RI.configure(height=0)

        self.sheet.enable_bindings()
        self.sheet.set_width_of_index_to_text()

        buttonFrame = ButtonFrame(self, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        loadButton = ttk.Button(buttonFrame, text="Load from CSV", command=self.loadCSV)
        loadButton.pack(side="left", padx=padding)

        self.sheet.pack(side="top", expand=True, fill="both")

    def reset(self):
        self.sheet.set_sheet_data(
            self.formatData(self.titration.knownSignals.knownSpectra)
        )

    def saveData(self):
        data = np.array(self.sheet.get_sheet_data(), dtype=object)
        data[data == ""] = "nan"
        data = data.astype(float)

        self.knownSpectra = ma.masked_invalid(data)
        self.spectraTitles = self.titration.contributors.outputNames
        self.signalTitles = self.titration.processedSignalTitlesStrings
        self.saved = True
        self.destroy()

    def loadCSV(self):
        fileType = tk.StringVar(self)
        filePath = fd.askopenfilename(
            master=self,
            title="Load from CSV",
            filetypes=[("All files", "*.*"), ("CSV files", "*.csv")],
            typevariable=fileType,
        )
        if filePath == "":
            return
        with open(filePath, encoding="utf-8-sig") as file:
            d = csv.Sniffer().sniff(file.readline() + file.readline())
            file.seek(0)
            data = np.array(list(csv.reader(file, dialect=d)))

        if data[0, 0] == "":
            # first row contains signal titles
            signalTitles = data[0, 1:]
            spectraTitles = data[1:, 0]
            spectra = data[1:, 1:].astype(float)
            for spectrumTitle, spectrum in zip(spectraTitles, spectra):
                if spectrumTitle in self.titration.contributors.outputNames:
                    for signalTitle, value in zip(signalTitles, spectrum):
                        if signalTitle in self.titration.processedSignalTitlesStrings:
                            self.sheet.set_cell_data(
                                np.where(
                                    self.titration.contributors.outputNames
                                    == spectrumTitle
                                )[0][0],
                                np.where(
                                    self.titration.processedSignalTitlesStrings
                                    == signalTitle
                                )[0][0],
                                value,
                            )
        else:
            spectraTitles = data[:, 0]
            spectra = data[:, 1:].astype(float)
            for spectrumTitle, spectrum in zip(spectraTitles, spectra):
                if spectrumTitle in self.titration.contributors.outputNames:
                    self.sheet.set_row_data(
                        np.where(
                            self.titration.contributors.outputNames == spectrumTitle
                        )[0][0],
                        list(spectrum),
                    )
        self.sheet.redraw()


class KnownSpectraPerMoleculePopup(moduleFrame.Popup):
    def __init__(self, titration, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.titration = titration
        self.title("Enter any known spectra")

        height = int(self.master.winfo_height() * 0.4)
        self.frame = ttk.Frame(self, height=height)
        self.frame.pack(expand=True, fill="both")

        buttonFrame = ButtonFrame(self.frame, self.reset, self.saveData, self.destroy)
        buttonFrame.pack(expand=False, fill="both", side="bottom")

        self.createNotebook()

    def createNotebook(self):
        self.notebook = ttk.Notebook(self.frame, style="Flat.TNotebook")
        self.notebook.pack(expand=True, fill="both", padx=padding, pady=padding)

        self.sheets = []

        splitIndices = np.cumsum(
            self.titration.contributors.contributorsCountPerMolecule
        )[:-1]

        spectraTitlesPerMolecule = np.split(
            self.titration.contributors.outputNames, splitIndices
        )
        knownSpectraPerMolecule = np.vsplit(
            self.titration.knownSignals.knownSpectra, splitIndices
        )
        signalsFilterPerMolecule = [
            self.titration.contributingSpecies.signalToMoleculeMap == i
            for i in range(self.titration.speciation.freeCount)
        ]

        for (
            molecule,
            spectraTitles,
            knownSpectra,
            contributorsCount,
            signalsFilter,
        ) in zip(
            self.titration.speciation.freeNames,
            spectraTitlesPerMolecule,
            knownSpectraPerMolecule,
            self.titration.contributors.contributorsCountPerMolecule,
            signalsFilterPerMolecule,
        ):
            if contributorsCount > 0:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    sheet = Sheet(
                        self,
                        empty_vertical=0,
                        empty_horizontal=0,
                        data=list(
                            knownSpectra[:, signalsFilter].astype(str).filled("")
                        ),
                        headers=list(
                            self.titration.processedSignalTitlesStrings[signalsFilter]
                        ),
                        row_index=list(spectraTitles),
                        set_all_heights_and_widths=True,
                    )
                sheet.MT.configure(height=sheet.MT.row_positions[-1] + 1)
                sheet.RI.configure(height=0)

                sheet.enable_bindings()
                sheet.set_width_of_index_to_text()

                self.notebook.add(sheet, text=molecule)
                self.sheets.append(sheet)
            else:
                self.sheets.append(None)

    def reset(self):
        # TODO: FIX!
        self.contributorsTable.destroy()
        self.createTable()

    def saveData(self):
        knownSpectraAll = ma.masked_all(
            [
                self.titration.contributors.outputCount,
                self.titration.processedSignalCount,
            ]
        )
        splitIndices = np.cumsum(
            self.titration.contributors.contributorsCountPerMolecule
        )[:-1]
        knownSpectraPerMolecule = np.vsplit(
            knownSpectraAll, splitIndices
        )  # returns a view, so can edit

        signalsFilterPerMolecule = [
            self.titration.contributingSpecies.signalToMoleculeMap == i
            for i in range(self.titration.speciation.freeCount)
        ]

        for sheet, knownSpectra, signalsFilter in zip(
            self.sheets,
            knownSpectraPerMolecule,
            signalsFilterPerMolecule,
        ):
            if sheet is None:
                continue
            data = np.array(sheet.get_sheet_data(), dtype=object)
            data[data == ""] = "nan"
            data = data.astype(float)
            knownSpectra[:, signalsFilter] = ma.masked_invalid(data)

        self.knownSpectra = knownSpectraAll
        self.spectraTitles = self.titration.contributors.outputNames
        self.signalTitles = self.titration.processedSignalTitlesStrings
        self.saved = True
        self.destroy()


class GetKnownSpectra(KnownSignals):
    Popup = KnownSpectraPopup
    popupAttributes = ("knownSpectra", "spectraTitles", "signalTitles")

    @property
    def Popup(self):
        singleMoleculeIndex = self.titration.contributingSpecies.singleMoleculeIndex
        if type(singleMoleculeIndex) is np.ndarray:
            return KnownSpectraPerMoleculePopup
        else:
            return KnownSpectraPopup

    @property
    def knownSpectra(self):
        currentContributors = self.titration.contributors.outputNames
        lastContributors = self.spectraTitles
        lastSignals = self.signalTitles
        currentSignals = self.titration.processedSignalTitlesStrings

        if not set(currentSignals).issubset(lastSignals):
            return ma.masked_all((len(currentContributors), len(currentSignals)))
        relevantSignalIndices = [
            np.where(lastSignals == signal)[0][0] for signal in currentSignals
        ]
        relevantSignals = self._knownSpectra[:, relevantSignalIndices]

        output = ma.masked_all((len(currentContributors), len(currentSignals)))
        for spectrumTitle, spectrum in zip(lastContributors, relevantSignals):
            if spectrumTitle in currentContributors:
                output[
                    np.where(currentContributors == spectrumTitle)[0][0], :
                ] = spectrum
        return output

    @knownSpectra.setter
    def knownSpectra(self, value):
        self._knownSpectra = value


class GetAllSpectra(KnownSignals):
    @property
    def knownSpectra(self):
        return ma.masked_all(
            (
                self.titration.contributors.outputCount,
                self.titration.processedSignalCount,
            )
        )


class ModuleFrame(moduleFrame.ModuleFrame):
    group = "Spectra"
    dropdownLabelText = "Specify any known spectra?"
    dropdownOptions = {
        "Optimise all spectra": GetAllSpectra,
        "Specify some known spectra": GetKnownSpectra,
    }
    attributeName = "knownSignals"
